#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Methods for calculating pure component properties from:

The Properties of Gases & Liquids, 5th Edition
Reid, Prausnitz and Polling, 2001, McGraw-Hill

All parameter indices based on conventions used by the source
"""
# TODO: Missing doc strings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from pyomo.environ import log, Var, units as pyunits

from idaes.core.util.misc import set_param_from_config

from idaes.core.util.constants import Constants as const


# -----------------------------------------------------------------------------
class RPP5(object):
    # -----------------------------------------------------------------------------
    # Heat capacities, enthalpies and entropies
    class cp_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.cp_mol_ig_comp_coeff_a0 = Var(
                doc="Parameter a0 for ideal gas molar heat capacity", units=None
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="a0")

            cobj.cp_mol_ig_comp_coeff_a1 = Var(
                doc="Parameter a1 for ideal gas molar heat capacity",
                units=pyunits.K**-1,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="a1")

            cobj.cp_mol_ig_comp_coeff_a2 = Var(
                doc="Parameter a2 for ideal gas molar heat capacity",
                units=pyunits.K**-2,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="a2")

            cobj.cp_mol_ig_comp_coeff_a3 = Var(
                doc="Parameter a3 for ideal gas molar heat capacity",
                units=pyunits.K**-3,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="a3")

            cobj.cp_mol_ig_comp_coeff_a4 = Var(
                doc="Parameter a4 for ideal gas molar heat capacity",
                units=pyunits.K**-4,
            )
            set_param_from_config(cobj, param="cp_mol_ig_comp_coeff", index="a4")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific heat capacity
            T = pyunits.convert(T, to_units=pyunits.K)

            cp = (
                cobj.cp_mol_ig_comp_coeff_a4 * T**4
                + cobj.cp_mol_ig_comp_coeff_a3 * T**3
                + cobj.cp_mol_ig_comp_coeff_a2 * T**2
                + cobj.cp_mol_ig_comp_coeff_a1 * T
                + cobj.cp_mol_ig_comp_coeff_a0
            ) * const.gas_constant

            units = b.params.get_metadata().derived_units
            return pyunits.convert(cp, units.HEAT_CAPACITY_MOLE)

    class enth_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff_a0"):
                RPP5.cp_mol_ig_comp.build_parameters(cobj)

            if cobj.parent_block().config.include_enthalpy_of_formation:
                units = cobj.parent_block().get_metadata().derived_units

                cobj.enth_mol_form_vap_comp_ref = Var(
                    doc="Vapor phase molar heat of formation @ Tref",
                    units=units.ENERGY_MOLE,
                )
                set_param_from_config(cobj, param="enth_mol_form_vap_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific enthalpy
            T = pyunits.convert(T, to_units=pyunits.K)
            Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

            units = b.params.get_metadata().derived_units

            h = (
                pyunits.convert(
                    (
                        (cobj.cp_mol_ig_comp_coeff_a4 / 5) * (T**5 - Tr**5)
                        + (cobj.cp_mol_ig_comp_coeff_a3 / 4) * (T**4 - Tr**4)
                        + (cobj.cp_mol_ig_comp_coeff_a2 / 3) * (T**3 - Tr**3)
                        + (cobj.cp_mol_ig_comp_coeff_a1 / 2) * (T**2 - Tr**2)
                        + cobj.cp_mol_ig_comp_coeff_a0 * (T - Tr)
                    )
                    * const.gas_constant,
                    units.ENERGY_MOLE,
                )
                + cobj.enth_mol_form_vap_comp_ref
            )

            return h

    class entr_mol_ig_comp:
        @staticmethod
        def build_parameters(cobj):
            if not hasattr(cobj, "cp_mol_ig_comp_coeff_a0"):
                RPP5.cp_mol_ig_comp.build_parameters(cobj)

            units = cobj.parent_block().get_metadata().derived_units

            cobj.entr_mol_form_vap_comp_ref = Var(
                doc="Vapor phase molar entropy of formation @ Tref",
                units=units.ENTROPY_MOLE,
            )
            set_param_from_config(cobj, param="entr_mol_form_vap_comp_ref")

        @staticmethod
        def return_expression(b, cobj, T):
            # Specific entropy
            T = pyunits.convert(T, to_units=pyunits.K)
            Tr = pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)

            units = b.params.get_metadata().derived_units

            s = (
                pyunits.convert(
                    (
                        (cobj.cp_mol_ig_comp_coeff_a4 / 4) * (T**4 - Tr**4)
                        + (cobj.cp_mol_ig_comp_coeff_a3 / 3) * (T**3 - Tr**3)
                        + (cobj.cp_mol_ig_comp_coeff_a2 / 2) * (T**2 - Tr**2)
                        + cobj.cp_mol_ig_comp_coeff_a1 * (T - Tr)
                        + cobj.cp_mol_ig_comp_coeff_a0 * log(T / Tr)
                    )
                    * const.gas_constant,
                    units.ENTROPY_MOLE,
                )
                + cobj.entr_mol_form_vap_comp_ref
            )

            return s

    # -----------------------------------------------------------------------------
    # Antoine equation for saturation pressure
    class pressure_sat_comp:
        @staticmethod
        def build_parameters(cobj):
            cobj.pressure_sat_comp_coeff_A = Var(
                doc="Antoine A coefficient for calculating Psat",
                units=pyunits.dimensionless,
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="A")

            cobj.pressure_sat_comp_coeff_B = Var(
                doc="Antoine B coefficient for calculating Psat", units=pyunits.K
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="B")

            cobj.pressure_sat_comp_coeff_C = Var(
                doc="Antoine C coefficient for calculating Psat", units=pyunits.K
            )
            set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="C")

        @staticmethod
        def return_expression(b, cobj, T, dT=False):
            if dT:
                return RPP5.pressure_sat_comp.dT_expression(b, cobj, T)

            psat = (
                10
                ** (
                    cobj.pressure_sat_comp_coeff_A
                    - (
                        cobj.pressure_sat_comp_coeff_B
                        / (T + cobj.pressure_sat_comp_coeff_C - 273.15 * pyunits.degK)
                    )
                )
                * pyunits.bar
            )

            base_units = b.params.get_metadata().default_units
            p_units = base_units.PRESSURE
            return pyunits.convert(psat, to_units=p_units)

        @staticmethod
        def dT_expression(b, cobj, T):
            p_sat_dT = (
                RPP5.pressure_sat_comp.return_expression(b, cobj, T)
                * cobj.pressure_sat_comp_coeff_B
                * log(10)
                / (T + cobj.pressure_sat_comp_coeff_C - 273.15 * pyunits.degK) ** 2
            )

            base_units = b.params.get_metadata().default_units
            dp_units = base_units.PRESSURE * base_units.TEMPERATURE**-1
            return pyunits.convert(p_sat_dT, to_units=dp_units)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

import pytest

from pyomo.util.check_units import assert_units_consistent
from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Objective,
    value,
    units as pyunits,
)

from idaes.core import FlowsheetBlock
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available
from idaes.core.util.scaling import constraint_scaling_transform
from idaes.models_extra.power_generation.properties.natural_gas_PR import get_prop
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.unit_models import GibbsReactor
from idaes.models.properties.modular_properties.eos.ceos import Cubic
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog

SOUT = idaeslog.INFO

# Set module level pyest marker
pytestmark = pytest.mark.cubic_root

# Get default solver for testing
solver = get_solver()


class TestNaturalGasProps(object):
    @pytest.fixture()
    def m(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = GenericParameterBlock(
            **get_prop(
                components=[
                    "H2",
                    "CO",
                    "H2O",
                    "CO2",
                    "CH4",
                    "C2H6",
                    "C3H8",
                    "C4H10",
                    "N2",
                    "O2",
                    "Ar",
                ]
            )
        )
        m.fs.state = m.fs.props.build_state_block([1], defined_state=True)

        return m

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.integration
    def test_T_sweep(self, m):
        assert_units_consistent(m)

        m.fs.obj = Objective(expr=(m.fs.state[1].temperature - 510) ** 2)

        for logP in range(8, 13, 1):
            m.fs.state[1].flow_mol.fix(100)
            m.fs.state[1].mole_frac_comp["H2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["CO"].fix(0.1)
            m.fs.state[1].mole_frac_comp["H2O"].fix(0.1)
            m.fs.state[1].mole_frac_comp["CO2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["O2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["N2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["Ar"].fix(0.1)
            m.fs.state[1].mole_frac_comp["CH4"].fix(0.1)
            m.fs.state[1].mole_frac_comp["C2H6"].fix(0.1)
            m.fs.state[1].mole_frac_comp["C3H8"].fix(0.05)
            m.fs.state[1].mole_frac_comp["C4H10"].fix(0.05)
            m.fs.state[1].temperature.fix(300)
            m.fs.state[1].pressure.fix(10 ** (0.5 * logP))

            m.fs.state.initialize()

            m.fs.state[1].temperature.unfix()

            results = solver.solve(m, tee=True)

            assert check_optimal_termination(results)
            assert -93000 == pytest.approx(
                value(m.fs.state[1].enth_mol_phase["Vap"]), 1
            )
            assert 250 == pytest.approx(value(m.fs.state[1].entr_mol_phase["Vap"]), 1)

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.integration
    def test_P_sweep(self, m):
        for T in range(300, 1000, 200):
            m.fs.state[1].flow_mol.fix(100)
            m.fs.state[1].mole_frac_comp["H2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["CO"].fix(0.1)
            m.fs.state[1].mole_frac_comp["H2O"].fix(0.1)
            m.fs.state[1].mole_frac_comp["CO2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["O2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["N2"].fix(0.1)
            m.fs.state[1].mole_frac_comp["Ar"].fix(0.1)
            m.fs.state[1].mole_frac_comp["CH4"].fix(0.1)
            m.fs.state[1].mole_frac_comp["C2H6"].fix(0.1)
            m.fs.state[1].mole_frac_comp["C3H8"].fix(0.05)
            m.fs.state[1].mole_frac_comp["C4H10"].fix(0.05)
            m.fs.state[1].temperature.fix(T)
            m.fs.state[1].pressure.fix(1e5)

            m.fs.state.initialize()

            results = solver.solve(m)

            assert check_optimal_termination(results)

            while m.fs.state[1].pressure.value <= 1e6:
                m.fs.state[1].pressure.value = m.fs.state[1].pressure.value + 2e5
                results = solver.solve(m)

                assert check_optimal_termination(results)

                assert -70000 >= value(m.fs.state[1].enth_mol_phase["Vap"])
                assert -105000 <= value(m.fs.state[1].enth_mol_phase["Vap"])
                assert 185 <= value(m.fs.state[1].entr_mol_phase["Vap"])
                assert 265 >= value(m.fs.state[1].entr_mol_phase["Vap"])

    @pytest.mark.skipif(
        not cubic_roots_available(), reason="Cubic functions not available"
    )
    @pytest.mark.component
    def test_gibbs(self, m):
        m.fs.props = GenericParameterBlock(
            **get_prop(components=["H2", "CO", "H2O", "CO2", "O2", "N2", "Ar", "CH4"])
        )
        m.fs.reactor = GibbsReactor(
            dynamic=False,
            has_heat_transfer=True,
            has_pressure_change=False,
            property_package=m.fs.props,
        )

        m.fs.reactor.inlet.flow_mol.fix(8000)  # mol/s
        m.fs.reactor.inlet.temperature.fix(600)  # K
        m.fs.reactor.inlet.pressure.fix(1e6)  # Pa

        m.fs.reactor.outlet.temperature.fix(600)  # K

        m.fs.reactor.inlet.mole_frac_comp[0, "H2"].fix(0)
        m.fs.reactor.inlet.mole_frac_comp[0, "CO"].fix(0)
        m.fs.reactor.inlet.mole_frac_comp[0, "H2O"].fix(0.3)
        m.fs.reactor.inlet.mole_frac_comp[0, "CO2"].fix(0)
        m.fs.reactor.inlet.mole_frac_comp[0, "O2"].fix(0.2)
        m.fs.reactor.inlet.mole_frac_comp[0, "N2"].fix(0.05)
        m.fs.reactor.inlet.mole_frac_comp[0, "Ar"].fix(0.05)
        m.fs.reactor.inlet.mole_frac_comp[0, "CH4"].fix(0.4)

        constraint_scaling_transform(
            m.fs.reactor.control_volume.enthalpy_balances[0.0], 1e-6
        )
        m.fs.reactor.gibbs_scaling = 1e-6

        results = solver.solve(m, tee=True)

        assert check_optimal_termination(results)

        assert 0.017 == pytest.approx(
            value(m.fs.reactor.outlet.mole_frac_comp[0, "H2"]), 1e-1
        )
        assert 0.0001 == pytest.approx(
            value(m.fs.reactor.outlet.mole_frac_comp[0, "CO"]), 1
        )
        assert 0.487 == pytest.approx(
            value(m.fs.reactor.outlet.mole_frac_comp[0, "H2O"]), 1e-1
        )
        assert 0.103 == pytest.approx(
            value(m.fs.reactor.outlet.mole_frac_comp[0, "CO2"]), 1e-1
        )
        assert 0.293 == pytest.approx(
            value(m.fs.reactor.outlet.mole_frac_comp[0, "CH4"]), 1e-1
        )

        assert -634e6 == pytest.approx(value(m.fs.reactor.heat_duty[0]), 1e-3)


# Reference: CoolProp: http://www.coolprop.org/index.html
data = {
    "phase": "Vap",
    "pressure": 101325,
    "temperature": 311.8730783132979,
    "enth_mol_phase": 24188.517506218643,
    "heat_capacity_ratio_phase": 1.290590129342867,
    "cp_mol_phase": 37.726122878482165,
    "cv_mol_phase": 29.2316840341025,
    "isentropic_speed_sound_phase": 279.2959607849875,
    "mole_frac_phase_comp": {"CO2": 0.94, "H2O": 0.06},
}


@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class Test_CO2_H2O_Properties:
    @pytest.fixture(scope="class")
    def build_model(self):
        m = ConcreteModel()

        # Properties
        comp_props = get_prop(components=["CO2", "H2O"], phases=["Vap", "Liq"])

        # Parameters block
        m.params = GenericParameterBlock(**comp_props)

        m.props = m.params.build_state_block(
            [1], defined_state=True, parameters=m.params, has_phase_equilibrium=True
        )

        m.props[1].flow_mol.fix(100)
        m.props[1].pressure.fix(101325)
        m.props[1].mole_frac_comp["CO2"].fix(0.94)
        m.props[1].mole_frac_comp["H2O"].fix(0.06)
        m.props[1].temperature.fix(311.8730783132979)

        # Trigger construction of intermediate variables and constraints
        tmp = m.props[1].cv_mol_phase["Vap"]

        assert degrees_of_freedom(m) == 0

        m.props.initialize()
        results = get_solver(options={"bound_push": 1e-8}).solve(m)

        assert check_optimal_termination(results)

        return m

    @pytest.mark.unit
    def test_cp_mol_phase(self, build_model):
        m = build_model
        assert (
            str(pyunits.get_units(Cubic.cp_mol_phase(m.props[1], "Vap")))
            == "kg*m**2/K/mol/s**2"
        )

        assert (
            pytest.approx(value(Cubic.cp_mol_phase(m.props[1], "Vap")), rel=0.1)
            == data["cp_mol_phase"]
        )

    @pytest.mark.unit
    def test_cv_mol_phase(self, build_model):
        m = build_model
        assert (
            str(pyunits.get_units(Cubic.cv_mol_phase(m.props[1], "Vap")))
            == "kg*m**2/K/mol/s**2"
        )

        assert (
            pytest.approx(value(Cubic.cv_mol_phase(m.props[1], "Vap")), rel=0.1)
            == data["cv_mol_phase"]
        )

    @pytest.mark.unit
    def test_heat_capacity_ratio_phase(self, build_model):
        m = build_model
        assert (
            str(pyunits.get_units(Cubic.heat_capacity_ratio_phase(m.props[1], "Vap")))
            == "dimensionless"
        )

        assert (
            pytest.approx(
                value(Cubic.heat_capacity_ratio_phase(m.props[1], "Vap")), rel=0.1
            )
            == data["heat_capacity_ratio_phase"]
        )

    @pytest.mark.unit
    def test_isentropic_speed_sound_phase(self, build_model):
        m = build_model
        assert (
            str(
                pyunits.get_units(Cubic.isentropic_speed_sound_phase(m.props[1], "Vap"))
            )
            == "m/s"
        )

        assert (
            pytest.approx(
                value(Cubic.isentropic_speed_sound_phase(m.props[1], "Vap")), rel=0.1
            )
            == data["isentropic_speed_sound_phase"]
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Waterwall section model test

main equations:

* Heat is given by fire-side boiler model
* Calculate pressure change due to friction and gravity
* Calculate slag layer wall temperature
* Consider a layer of metal and a layer of slag


Created on Thu Aug 24 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest

# Import Pyomo libraries
import pyomo.environ as pyo
from pyomo.network import Arc

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom


# Import Unit Model Modules
from idaes.models.properties import iapws95

from idaes.models_extra.power_generation.unit_models.waterwall_section import (
    WaterwallSection,
)
from idaes.core.solvers import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def model():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.prop_water = iapws95.Iapws95ParameterBlock()
    n_waterwalls = 10
    m.fs.ww_zones = pyo.RangeSet(n_waterwalls)
    m.fs.Waterwalls = WaterwallSection(
        m.fs.ww_zones,
        dynamic=False,
        has_holdup=False,
        property_package=m.fs.prop_water,
        has_heat_transfer=True,
        has_pressure_change=True,
    )

    def arc_rule(b, i):
        return {
            "source": m.fs.Waterwalls[i].outlet,
            "destination": m.fs.Waterwalls[i + 1].inlet,
        }

    m.arc = Arc(pyo.RangeSet(n_waterwalls - 1), rule=arc_rule)

    # Pyomo expands arcs writing constraints outlet unit 1 = inlet unit 2
    pyo.TransformationFactory("network.expand_arcs").apply_to(m)

    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.unit
def test_basic_build(model):
    """Make a turbine model and make sure it doesn't throw exception"""
    assert degrees_of_freedom(model) == 103
    # Check unit config arguments
    assert len(model.fs.Waterwalls[1].config) == 10
    assert model.fs.Waterwalls[1].config.has_heat_transfer
    assert model.fs.Waterwalls[1].config.has_pressure_change
    assert model.fs.Waterwalls[1].config.property_package is model.fs.prop_water


# @pytest.mark.integration
# def test_units(model):
#     assert_units_consistent(model)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_waterwall(model):
    # fix inputs
    # 10 waterwall sections
    for i in model.fs.ww_zones:
        model.fs.Waterwalls[i].tube_diameter.fix(0.047)
        model.fs.Waterwalls[i].tube_thickness.fix(0.00350)
        model.fs.Waterwalls[i].fin_thickness.fix(0.00455)
        model.fs.Waterwalls[i].slag_thickness[:].fix(0.001)
        model.fs.Waterwalls[i].fin_length.fix(0.0115)
        model.fs.Waterwalls[i].number_tubes.fix(610)
        model.fs.Waterwalls[i].fcorrection_dp.fix(1.2)

    # water wall section height (must be equal to the fire side model zones)
    model.fs.Waterwalls[1].height.fix(6.150)
    model.fs.Waterwalls[2].height.fix(3.150)
    model.fs.Waterwalls[3].height.fix(1.5)
    model.fs.Waterwalls[4].height.fix(1.450)
    model.fs.Waterwalls[5].height.fix(1.350)
    model.fs.Waterwalls[6].height.fix(1.250)
    model.fs.Waterwalls[7].height.fix(1.150)
    model.fs.Waterwalls[8].height.fix(1.350)
    model.fs.Waterwalls[9].height.fix(3.250)
    model.fs.Waterwalls[10].height.fix(3.450)

    # water wall section projected area
    model.fs.Waterwalls[1].projected_area.fix(320.0)
    model.fs.Waterwalls[2].projected_area.fix(150.3)
    model.fs.Waterwalls[3].projected_area.fix(70.8)
    model.fs.Waterwalls[4].projected_area.fix(70.0)
    model.fs.Waterwalls[5].projected_area.fix(58.6)
    model.fs.Waterwalls[6].projected_area.fix(58.6)
    model.fs.Waterwalls[7].projected_area.fix(50.1)
    model.fs.Waterwalls[8].projected_area.fix(65.6)
    model.fs.Waterwalls[9].projected_area.fix(145.6)
    model.fs.Waterwalls[10].projected_area.fix(165.5)

    # Heat loss to waterwall Q in W
    model.fs.Waterwalls[1].heat_fireside[:].fix(2.3e7)
    model.fs.Waterwalls[2].heat_fireside[:].fix(1.5e7)
    model.fs.Waterwalls[3].heat_fireside[:].fix(6.9e6)
    model.fs.Waterwalls[4].heat_fireside[:].fix(1.2e7)
    model.fs.Waterwalls[5].heat_fireside[:].fix(1.2e7)
    model.fs.Waterwalls[6].heat_fireside[:].fix(1.2e7)
    model.fs.Waterwalls[7].heat_fireside[:].fix(1.1e7)
    model.fs.Waterwalls[8].heat_fireside[:].fix(9.9e6)
    model.fs.Waterwalls[9].heat_fireside[:].fix(2.2e7)
    model.fs.Waterwalls[10].heat_fireside[:].fix(1.9e7)

    optarg = {"tol": 1e-7, "linear_solver": "ma27", "max_iter": 40}
    solver.options = optarg

    # Set inlet and operating conditions, and some initial conditions.
    model.fs.Waterwalls[1].inlet.flow_mol[0].fix(150055.0)  # mol/s
    model.fs.Waterwalls[1].inlet.enth_mol[0].fix(31000.0)  # J/mol
    model.fs.Waterwalls[1].inlet.pressure[0].fix(1.750e7)  # Pa

    model.fs.Waterwalls[1].initialize(
        state_args={
            "flow_mol": model.fs.Waterwalls[1].inlet.flow_mol[0].value,
            "pressure": model.fs.Waterwalls[1].inlet.pressure[0].value,
            "enth_mol": model.fs.Waterwalls[1].inlet.enth_mol[0].value,
        },
        optarg=optarg,
    )

    for i in range(2, 11):
        model.fs.Waterwalls[i].initialize(
            state_args={
                "flow_mol": model.fs.Waterwalls[i - 1].outlet.flow_mol[0].value,
                "pressure": model.fs.Waterwalls[i - 1].outlet.pressure[0].value,
                "enth_mol": model.fs.Waterwalls[i - 1].outlet.enth_mol[0].value,
            },
            optarg=optarg,
        )

    assert degrees_of_freedom(model) == 0  # only Waterwalls[1] input is fixed


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_waterwall(model):
    results = solver.solve(model, tee=True)
    # test energy balance
    heat_duty = sum(
        pyo.value(model.fs.Waterwalls[i].heat_duty[0]) for i in range(1, 11)
    )
    Fhin = (
        model.fs.Waterwalls[1].control_volume.properties_in[0].flow_mol
        * model.fs.Waterwalls[1].control_volume.properties_in[0].enth_mol
    )
    Fhout = (
        model.fs.Waterwalls[10].control_volume.properties_out[0].flow_mol
        * model.fs.Waterwalls[10].control_volume.properties_out[0].enth_mol
    )
    assert pytest.approx(heat_duty, abs=1e-3) == pyo.value(Fhout - Fhin)
    assert pytest.approx(0.08353, abs=1e-3) == pyo.value(
        model.fs.Waterwalls[10].control_volume.properties_out[0].vapor_frac
    )
    assert pytest.approx(150055.0, abs=1e-3) == pyo.value(
        model.fs.Waterwalls[10].control_volume.properties_out[0].flow_mol
    )
    # test mass conservation
    assert pytest.approx(0, abs=1e-3) == pyo.value(
        model.fs.Waterwalls[10].control_volume.properties_out[0].flow_mol
        - model.fs.Waterwalls[1].control_volume.properties_in[0].flow_mol
    )
    assert pytest.approx(31951.65106, abs=1e-3) == pyo.value(
        model.fs.Waterwalls[10].control_volume.properties_out[0].enth_mol
    )
    assert degrees_of_freedom(model) == 0
    # Check for optimal solution
    assert pyo.check_optimal_termination(results)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Property package for the reaction of CH4 with an iron-based OC.
Overall reducer reactions for Methane combustion:
(1) CH4 + 12Fe2O3 => 8Fe3O4 + CO2 + 2H2O

Parameters and equations written in this model were primarily derived from:
A. Abad, J. Adánez, F. García-Labiano, L.F. de Diego, P. Gayán, J. Celaya,
Mapping of the range of operational conditions for cu-, Fe-, and Ni-based
oxygen carriers in chemical-looping combustion,
Chem. Eng. Sci. 62 (2007) 533–549.

"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# Import Pyomo libraries
from pyomo.environ import (
    Constraint,
    exp,
    Param,
    Reals,
    Set,
    value,
    Var,
    units as pyunits,
)
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.common.config import ConfigBlock, ConfigValue, Bool


# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    ReactionParameterBlock,
    ReactionBlockDataBase,
    ReactionBlockBase,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.initialization import (
    fix_state_vars,
    revert_state_vars,
    solve_indexed_blocks,
)
from idaes.core.util.model_statistics import (
    number_unfixed_variables_in_activated_equalities,
)
from idaes.core.util.config import (
    is_state_block,
    is_physical_parameter_block,
    is_reaction_parameter_block,
)
from idaes.core.util.constants import Constants
import idaes.logger as idaeslog
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

# Some more information about this module
__author__ = "Chinedu Okoli"


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("HeteroReactionParameterBlock")
class ReactionParameterData(ReactionParameterBlock):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """

    # Create Class ConfigBlock
    CONFIG = ConfigBlock()
    CONFIG.declare(
        "gas_property_package",
        ConfigValue(
            description="Reference to associated PropertyPackageParameter "
            "object for the gas phase.",
            domain=is_physical_parameter_block,
        ),
    )
    CONFIG.declare(
        "solid_property_package",
        ConfigValue(
            description="Reference to associated PropertyPackageParameter "
            "object for the solid phase.",
            domain=is_physical_parameter_block,
        ),
    )
    CONFIG.declare(
        "default_arguments",
        ConfigBlock(
            description="Default arguments to use with Property Package", implicit=True
        ),
    )

    def build(self):
        """
        Callable method for Block construction.
        """
        super(ReactionParameterBlock, self).build()

        self._reaction_block_class = ReactionBlock

        # Reaction Index
        self.rate_reaction_idx = Set(initialize=["R1"])

        # Smoothing factor
        self.eps = Param(
            mutable=True,
            default=1e-8,
            doc="Smoothing Factor",
            units=pyunits.mol / pyunits.m**3,
        )
        # Reaction rate scale factor
        self._scale_factor_rxn = Param(
            mutable=True,
            default=1,
            doc="Scale Factor for reaction eqn." "Used to help initialization routine",
        )

        # Reaction Stoichiometry
        self.rate_reaction_stoichiometry = {
            ("R1", "Vap", "CH4"): -1,
            ("R1", "Vap", "CO2"): 1,
            ("R1", "Vap", "H2O"): 2,
            ("R1", "Sol", "Fe2O3"): -12,
            ("R1", "Sol", "Fe3O4"): 8,
            ("R1", "Sol", "Al2O3"): 0,
        }

        # Standard Heat of Reaction - J/mol_rxn
        dh_rxn_dict = {"R1": 136.5843e3}
        # Heat of Reaction should be defined in default units
        # (default energy units)/(default amount units)
        # per the define_meta.add_default_units method in the
        # relevant phase - in this case, the solid phase properties
        self.dh_rxn = Param(
            self.rate_reaction_idx,
            initialize=dh_rxn_dict,
            doc="Heat of reaction [J/mol]",
            units=pyunits.J / pyunits.mol,
        )

        # -------------------------------------------------------------------------
        # Reaction properties that can be estimated

        # Particle grain radius within OC particle
        self.grain_radius = Var(
            domain=Reals,
            initialize=2.6e-7,
            doc="Representative particle grain" "radius within OC particle [m]",
            units=pyunits.m,
        )
        self.grain_radius.fix()

        # Molar density OC particle
        self.dens_mol_sol = Var(
            domain=Reals,
            initialize=32811,
            doc="Molar density of OC particle [mol/m^3]",
            units=pyunits.mol / pyunits.m**3,
        )
        self.dens_mol_sol.fix()

        # Available volume for reaction - from EPAT report (1-ep)'
        self.a_vol = Var(
            domain=Reals,
            initialize=0.28,
            doc="Available reaction vol. per vol. of OC",
            units=pyunits.m**3 / pyunits.m**3,
        )
        self.a_vol.fix()

        # Activation Energy
        self.energy_activation = Var(
            self.rate_reaction_idx,
            domain=Reals,
            initialize=4.9e4,
            doc="Activation energy [J/mol]",
            units=pyunits.J / pyunits.mol,
        )
        self.energy_activation.fix()

        # Reaction order
        self.rxn_order = Var(
            self.rate_reaction_idx,
            domain=Reals,
            initialize=1.3,
            doc="Reaction order in gas species [-]",
        )
        self.rxn_order.fix()

        # Pre-exponential factor
        self.k0_rxn = Var(
            self.rate_reaction_idx,
            domain=Reals,
            initialize=8e-4,
            doc="Pre-exponential factor" "[mol^(1-rxn_order) * m^(3*rxn_order -2)/s]",
            units=pyunits.mol ** (1 - 1.3) * pyunits.m ** (3 * 1.3 - 2) / pyunits.s,
        )
        self.k0_rxn.fix()

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {
                "OC_conv": {"method": "_OC_conv", "units": None},
                "OC_conv_temp": {"method": "_OC_conv_temp", "units": None},
            }
        )
        obj.add_properties(
            {
                "k_rxn": {
                    "method": "_k_rxn",
                    "units": "mol^(1-rxn_order)m^(3*rxn_order -2)/s]",
                },
                "reaction_rate": {"method": "_reaction_rate", "units": "mol_rxn/m3.s"},
            }
        )

        obj.add_default_units(
            {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.kg,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


class _ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """

    def initialize(blk, outlvl=idaeslog.NOTSET, optarg=None, solver=None):
        """
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)
        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="reactions")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="reactions")

        init_log.info_high("Starting initialization")

        # TODO - Update in the future as needed
        # Get a single representative block for getting config arguments
        for k in blk.keys():
            rep_key = k
            break

        # Fix state variables if not already fixed
        # Fix state variables of the primary (solid) state block
        state_var_flags = fix_state_vars(blk[rep_key].config.solid_state_block)

        # Fix values of secondary (gas) state block variables if not fixed,
        # as well as the solid density variable.
        # This is done to keep the initialization problem square
        Cflag = {}  # Gas concentration flag
        Dflag = {}  # Solid density flag

        for k, b in blk.items():
            for j in b.gas_state_ref.params.component_list:
                if b.gas_state_ref.dens_mol_comp[j].fixed is True:
                    Cflag[k, j] = True
                else:
                    Cflag[k, j] = False
                    b.gas_state_ref.dens_mol_comp[j].fix(
                        b.gas_state_ref.dens_mol_comp[j].value
                    )
            if b.solid_state_ref.dens_mass_skeletal.fixed is True:
                Dflag[k] = True
            else:
                Dflag[k] = False
                b.solid_state_ref.dens_mass_skeletal.fix(
                    b.solid_state_ref.dens_mass_skeletal.value
                )

        # Initialize values
        for k, b in blk.items():
            if hasattr(b, "OC_conv_eqn"):
                calculate_variable_from_constraint(b.OC_conv, b.OC_conv_eqn)

            if hasattr(b, "OC_conv_temp_eqn"):
                calculate_variable_from_constraint(b.OC_conv_temp, b.OC_conv_temp_eqn)

            for j in b.params.rate_reaction_idx:
                if hasattr(b, "rate_constant_eqn"):
                    calculate_variable_from_constraint(
                        b.k_rxn[j], b.rate_constant_eqn[j]
                    )

                if hasattr(b, "gen_rate_expression"):
                    calculate_variable_from_constraint(
                        b.reaction_rate[j], b.gen_rate_expression[j]
                    )

        # Solve property block if non-empty
        free_vars = 0
        for k in blk.values():
            free_vars += number_unfixed_variables_in_activated_equalities(k)

        if free_vars > 0:
            # Create solver
            opt = get_solver(solver, optarg)
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                res = solve_indexed_blocks(opt, [blk], tee=slc.tee)
        else:
            res = ""
        init_log.info_high(
            "reactions initialization complete {}.".format(idaeslog.condition(res))
        )

        # ---------------------------------------------------------------------
        # Revert state vars and other variables to pre-initialization states
        # Revert state variables of the primary (solid) state block
        # TODO: I am not sure this line works as expected if there is more than one
        # time point
        for k in blk.values():
            revert_state_vars(k.config.solid_state_block, state_var_flags)

        for k, b in blk.items():
            for j in b.gas_state_ref.params.component_list:
                if Cflag[k, j] is False:
                    b.gas_state_ref.dens_mol_comp[j].unfix()
            if Dflag[k] is False:
                b.solid_state_ref.dens_mass_skeletal.unfix()

        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="reactions")
        init_log.info_high("States released.")


@declare_process_block_class("ReactionBlock", block_class=_ReactionBlock)
class ReactionBlockData(ReactionBlockDataBase):
    """
    Heterogeneous reaction package for methane reacting with Fe2O3 based OC
    """

    # Create Class ConfigBlock
    CONFIG = ConfigBlock()
    CONFIG.declare(
        "parameters",
        ConfigValue(
            domain=is_reaction_parameter_block,
            description="""
            A reference to an instance of the Reaction Parameter
            Block associated with this property package.
            """,
        ),
    )
    CONFIG.declare(
        "solid_state_block",
        ConfigValue(
            domain=is_state_block,
            description="""
            A reference to an instance of a StateBlock for the
            solid phase with which this reaction block should be associated.
            """,
        ),
    )
    CONFIG.declare(
        "gas_state_block",
        ConfigValue(
            domain=is_state_block,
            description="""
            A reference to an instance of a StateBlock for the
            gas phase with which this reaction block should be associated.
            """,
        ),
    )
    CONFIG.declare(
        "has_equilibrium",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Equilibrium reaction construction flag",
            doc="""
        Indicates whether terms for equilibrium controlled reactions
        should be constructed,
        **default** - True.
        **Valid values:** {
        **True** - include equilibrium reaction terms,
        **False** - exclude equilibrium reaction terms.}
        """,
        ),
    )

    def build(self):
        """
        Callable method for Block construction
        """
        super(ReactionBlockDataBase, self).build()

        # Object references to the corresponding state blocks and parameters
        add_object_reference(self, "_params", self.config.parameters)

        add_object_reference(
            self, "solid_state_ref", self.config.solid_state_block[self.index()]
        )
        add_object_reference(
            self, "gas_state_ref", self.config.gas_state_block[self.index()]
        )

        # Object reference for parameters if needed by CV1D
        # Reaction stoichiometry
        add_object_reference(
            self,
            "rate_reaction_stoichiometry",
            self.config.parameters.rate_reaction_stoichiometry,
        )

        # Heat of reaction
        add_object_reference(self, "dh_rxn", self.config.parameters.dh_rxn)

    # Rate constant method
    def _k_rxn(self):
        self.k_rxn = Var(
            self.params.rate_reaction_idx,
            domain=Reals,
            initialize=1,
            doc="Rate constant " "[mol^(1-rxn_order) * m^(3*rxn_order -2)/s]",
            units=pyunits.mol ** (1 - 1.3) * pyunits.m ** (3 * 1.3 - 2) / pyunits.s,
        )

        def rate_constant_eqn(b, j):
            if j == "R1":
                return self.k_rxn[j] == (
                    self.params.k0_rxn[j]
                    * exp(
                        -self.params.energy_activation[j]
                        / (
                            pyunits.convert(
                                Constants.gas_constant,  # J/mol/K
                                to_units=pyunits.J / pyunits.mol / pyunits.K,
                            )
                            * self.solid_state_ref.temperature
                        )
                    )
                )
            else:
                return Constraint.Skip

        try:
            # Try to build constraint
            self.rate_constant_eqn = Constraint(
                self.params.rate_reaction_idx, rule=rate_constant_eqn
            )
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.k_rxn)
            self.del_component(self.rate_constant_eqn)
            raise

    # Conversion of oxygen carrier
    def _OC_conv(self):
        self.OC_conv = Var(
            domain=Reals,
            initialize=0.0,
            doc="Fraction of metal oxide converted",
            units=pyunits.dimensionless,
        )

        def OC_conv_eqn(b):
            return (
                b.OC_conv
                * (
                    b.solid_state_ref.mass_frac_comp["Fe3O4"]
                    + (
                        b.solid_state_ref.params.mw_comp["Fe3O4"]
                        / b.solid_state_ref.params.mw_comp["Fe2O3"]
                    )
                    * (
                        b.params.rate_reaction_stoichiometry["R1", "Sol", "Fe3O4"]
                        / -b.params.rate_reaction_stoichiometry["R1", "Sol", "Fe2O3"]
                    )
                    * b.solid_state_ref.mass_frac_comp["Fe2O3"]
                )
                == b.solid_state_ref.mass_frac_comp["Fe3O4"]
            )

        try:
            # Try to build constraint
            self.OC_conv_eqn = Constraint(rule=OC_conv_eqn)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.OC_conv)
            self.del_component(self.OC_conv_eqn)

    # Conversion of oxygen carrier reformulated
    def _OC_conv_temp(self):
        self.OC_conv_temp = Var(
            domain=Reals,
            initialize=1.0,
            doc="Reformulation term for" "X to help eqn scaling",
            units=pyunits.dimensionless,
        )

        def OC_conv_temp_eqn(b):
            return b.OC_conv_temp**3 == (1 - b.OC_conv) ** 2

        try:
            # Try to build constraint
            self.OC_conv_temp_eqn = Constraint(rule=OC_conv_temp_eqn)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.OC_conv_temp)
            self.del_component(self.OC_conv_temp_eqn)

    # General rate of reaction method
    def _reaction_rate(self):
        self.reaction_rate = Var(
            self.params.rate_reaction_idx,
            domain=Reals,
            initialize=0,
            doc="Gen. rate of reaction [mol_rxn/m3.s]",
            units=pyunits.mol / pyunits.m**3 / pyunits.s,
        )

        def rate_rule(b, r):
            return b.reaction_rate[
                r
            ] == b.params._scale_factor_rxn * (  # pylint: disable=protected-access
                b.solid_state_ref.mass_frac_comp["Fe2O3"]
                * (1 - b.solid_state_ref.particle_porosity)
                * b.solid_state_ref.dens_mass_skeletal
                * (b.params.a_vol / (b.solid_state_ref.params.mw_comp["Fe2O3"]))
                * 3
                * -b.params.rate_reaction_stoichiometry["R1", "Sol", "Fe2O3"]
                * b.k_rxn[r]
                * (
                    (
                        (b.gas_state_ref.dens_mol_comp["CH4"] ** 2 + b.params.eps**2)
                        ** 0.5
                    )
                    ** b.params.rxn_order[r]
                )
                * b.OC_conv_temp
                / (b.params.dens_mol_sol * b.params.grain_radius)
                / (-b.params.rate_reaction_stoichiometry["R1", "Sol", "Fe2O3"])
            )

        try:
            # Try to build constraint
            self.gen_rate_expression = Constraint(
                self.params.rate_reaction_idx, rule=rate_rule
            )
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.reaction_rate)
            self.del_component(self.gen_rate_expression)
            raise

    def get_reaction_rate_basis(b):
        return MaterialFlowBasis.molar

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            _log.error("{} Temperature set below lower bound.".format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            _log.error("{} Temperature set above upper bound.".format(blk.name))

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # Set default scaling
        def _set_default_factor(v, s):
            for i in v:
                if iscale.get_scaling_factor(v[i]) is None:
                    iscale.set_scaling_factor(v[i], s)

        _set_default_factor(self.k_rxn, 1e3)
        _set_default_factor(self.OC_conv, 1e3)
        _set_default_factor(self.OC_conv_temp, 1e3)
        _set_default_factor(self.reaction_rate, 1e-2)

        if self.is_property_constructed("OC_conv_eqn"):
            iscale.constraint_scaling_transform(
                self.OC_conv_eqn,
                iscale.get_scaling_factor(self.OC_conv),
                overwrite=False,
            )

        if self.is_property_constructed("OC_conv_temp_eqn"):
            iscale.constraint_scaling_transform(
                self.OC_conv_temp_eqn,
                iscale.get_scaling_factor(self.OC_conv_temp),
                overwrite=False,
            )

        if self.is_property_constructed("rate_constant_eqn"):
            for i, c in self.rate_constant_eqn.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.k_rxn[i]), overwrite=False
                )

        if self.is_property_constructed("gen_rate_expression"):
            for i, c in self.gen_rate_expression.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.reaction_rate[i]), overwrite=False
                )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pyomo.environ as pyo
import pytest
import idaes
import idaes.core.solvers as isolve
import idaes.core.solvers.ipopt_l1


@pytest.mark.skipif(not pyo.SolverFactory("ipopt").available(False), reason="no Ipopt")
@pytest.mark.unit
def test_ipopt_idaes_config():
    """
    Test that the default solver options are set
    """
    with idaes.temporary_config_ctx():
        # in this context use idaes default solver options
        isolve.use_idaes_solver_configuration_defaults()
        idaes.cfg.ipopt.options.nlp_scaling_method = "toast-based"
        solver = pyo.SolverFactory("ipopt")
        assert solver.options["nlp_scaling_method"] == "toast-based"
        solver = pyo.SolverFactory("ipopt", options={"tol": 1})
        assert solver.options["tol"] == 1
        idaes.cfg.ipopt.options.tol = 1
        solver = pyo.SolverFactory("ipopt")
        assert solver.options["tol"] == 1


@pytest.mark.skipif(
    not pyo.SolverFactory("ipopt_l1").available(False), reason="no Ipopt_l1"
)
@pytest.mark.unit
def test_ipopt_l1_idaes_config():
    """
    Test that the default solver options are set
    """
    with idaes.temporary_config_ctx():
        # in this context use idaes default solver options
        isolve.use_idaes_solver_configuration_defaults()
        idaes.cfg.ipopt_l1.options.nlp_scaling_method = "toast-based"
        solver = pyo.SolverFactory("ipopt_l1")
        assert solver.options["nlp_scaling_method"] == "toast-based"
        solver = pyo.SolverFactory("ipopt_l1", options={"tol": 1})
        assert solver.options["tol"] == 1
        idaes.cfg.ipopt_l1.options.tol = 1
        solver = pyo.SolverFactory("ipopt_l1")
        assert solver.options["tol"] == 1


@pytest.mark.skipif(not pyo.SolverFactory("ipopt").available(False), reason="no Ipopt")
@pytest.mark.unit
def test_default_solver():
    """Test that default solver returns the correct solver type"""
    assert type(pyo.SolverFactory("default")) == type(
        pyo.SolverFactory(idaes.cfg.default_solver)
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for versioning
"""
# third-party
import pytest

# pkg
import idaes
from idaes import ver


@pytest.mark.unit
def test_idaes_version():
    assert idaes.__version__


@pytest.mark.unit
def test_ver_class():
    v = ver.Version(1, 2, 3)
    assert str(v) == "1.2.3"
    seq = [1, 2, 3]
    for i, n in enumerate(v):
        print("i:", i, "n:", n)
        assert n == seq[i]
    v = ver.Version(1, 2, 3, "beta", 1)
    assert str(v) == "1.2.3.b1"
    v = ver.Version(1, 2, 3, "beta", 2, "test")
    assert str(v) == "1.2.3.b2+test"
    v = ver.Version(1, 2, 3, "beta", label="test")
    assert str(v) == "1.2.3.b+test"
    v = ver.Version(1, 2, 3, "development")
    assert str(v) == "1.2.3.dev"
    pytest.raises(ValueError, ver.Version, 1, 2, 3, "howdy")


class MyVersionedClass(ver.HasVersion):
    def __init__(self):
        super(MyVersionedClass, self).__init__(1, 2, 3)


@pytest.mark.unit
def test_has_version():
    x = MyVersionedClass()
    assert str(x.version) == "1.2.3"


@pytest.mark.unit
def test_bump_version():
    v = ver.Version(1, 2, 3)
    assert tuple(v) == (1, 2, 3)
    v.micro += 1
    assert tuple(v) == (1, 2, 4)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base for IDAES process model objects.
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

# TODO: Look into if this is necessary
# pylint: disable=protected-access

import sys
import logging
import textwrap
from enum import Enum

from pandas import DataFrame

from pyomo.core.base.block import BlockData
from pyomo.common.formatting import tabular_writer
from pyomo.environ import Block
from pyomo.gdp import Disjunct
from pyomo.common.config import ConfigBlock

from idaes.core.base.process_block import declare_process_block_class
from idaes.core.initialization import BlockTriangularizationInitializer
from idaes.core.util.exceptions import (
    ConfigurationError,
    DynamicError,
    PropertyPackageError,
)
from idaes.core.util.tables import stream_table_dataframe_to_string
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_activated_constraints,
    number_activated_blocks,
)
from idaes.core.util.units_of_measurement import report_quantity


# Some more information about this module
__author__ = "John Eslick, Qi Chen, Andrew Lee"


__all__ = ["ProcessBlockData"]


useDefault = object()


# Set up logger
_log = logging.getLogger(__name__)


# Enumerate options for material flow basis
class MaterialFlowBasis(Enum):
    """
    Material flow basis Enum
    """

    molar = 0
    mass = 1
    other = 2


@declare_process_block_class("ProcessBaseBlock")
class ProcessBlockData(BlockData):
    """
    Base class for most IDAES process models and classes.

    The primary purpose of this class is to create the local config block to
    handle arguments provided by the user when constructing an object and to
    ensure that these arguments are stored in the config block.

    Additionally, this class contains a number of methods common to all IDAES
    classes.
    """

    CONFIG = ConfigBlock("ProcessBlockData", implicit=False)

    # Set default initializer and scaler
    default_initializer = BlockTriangularizationInitializer
    default_scaler = None

    def __init__(self, component):
        """
        Initialize a ProcessBlockData object.

        Args:
            component(Block): container Block instance to which this BlockData
                              belongs.

        Returns:
            (ProcessBlockData): A new instance
        """
        super(ProcessBlockData, self).__init__(component=component)
        self._pb_configured = False

    def build(self):
        """
        The build method is called by the default ProcessBlock rule.  If a rule
        is specified other than the default it is important to call
        ProcessBlockData's build method to put information from the "default"
        and "initialize" arguments to a ProcessBlock derived class into the
        BlockData object's ConfigBlock.

        The build method should usually be overloaded in a subclass derived
        from ProcessBlockData. This method would generally add Pyomo components
        such as variables, expressions, and constraints to the object. It is
        important for build() methods implemented in derived classes to call
        build() from the super class.

        Args:
            None

        Returns:
            None
        """
        self._get_config_args()

        # Add initialization order list, and populate with current model
        self.initialization_order = [self]

        # This is a dict to store default property scaling factors. They are
        # defined in the parameter block to provide a universal default for
        # quantities in a particular kind of state block.  For example, you can
        # set flow scaling once instead of for every state block. Some of these
        # may be left for the user to set and some may be defined in a property
        # module where reasonable defaults can be defined a priori. See
        # set_default_scaling, get_default_scaling, and unset_default_scaling
        self._default_scaling_factors = {}

    @property
    def default_scaling_factors(self):
        """
        Dict of default scaling factors for components in model
        """
        return self._default_scaling_factors

    @property
    def default_scaling_factor(self):
        # Backwards compatibility for name change
        # TODO: Deprecate once v2.0 is released
        return self._default_scaling_factors

    def set_default_scaling(self, attribute: str, value: float, index: str = None):
        """Set a default scaling factor for a property.

        Args:
            attribute: property attribute name
            value: default scaling factor
            index: for indexed properties, if this is not provided the default
                scaling factor applies to all indexed elements where specific
                indexes are not specified.

        Returns:
            None
        """
        self._default_scaling_factors[(attribute, index)] = value

    def unset_default_scaling(self, attribute: str, index: str = None):
        """Remove a previously set default value

        Args:
            attribute: property attribute name
            index: optional index for indexed properties

        Returns:
            None
        """
        try:
            del self._default_scaling_factors[(attribute, index)]
        except KeyError:
            pass

    def get_default_scaling(self, attribute: str, index: str = None):
        """Returns a default scale factor for a property

        Args:
            attribute: property attribute name
            index: optional index for indexed properties

        Returns:
            None
        """
        try:
            # If a specific component data index exists
            return self._default_scaling_factors[(attribute, index)]
        except KeyError:
            try:
                # indexed, but no specific index?
                return self._default_scaling_factors[(attribute, None)]
            except KeyError:
                # Can't find a default scale factor for what you asked for
                return None

    def flowsheet(self):
        """
        This method returns the components parent flowsheet object, i.e. the
        flowsheet component to which the model is attached. If the component
        has no parent flowsheet, the method returns None.

        Args:
            None

        Returns:
            Flowsheet object or None
        """
        parent = self.parent_block()

        while True:
            if parent is None:
                return None

            if hasattr(parent, "is_flowsheet") and parent.is_flowsheet():
                return parent

            else:
                parent = parent.parent_block()

    def _get_config_args(self):
        """
        Get config arguments for this element and put them in the ConfigBlock
        """
        if self._pb_configured:
            return
        self._pb_configured = True
        idx_map = self.parent_component()._idx_map  # index map function
        try:
            idx = self.index()
        except AttributeError:
            idx = None
        if idx_map is not None:
            idx = idx_map(idx)
        initialize = self.parent_component()._block_data_config_initialize
        if idx in initialize:
            kwargs = initialize[idx]
        else:
            kwargs = self.parent_component()._block_data_config_default
        self.config = self.CONFIG(kwargs)

    def fix_initial_conditions(self, state="steady-state"):
        """This method fixes the initial conditions for dynamic models.

        Args:
            state : initial state to use for simulation (default =
                    'steady-state')

        Returns :
            None
        """
        if state == "steady-state":
            for obj in self.component_objects((Block, Disjunct), descend_into=True):
                # Try to fix material_accumulation @ first time point
                try:
                    obj.material_accumulation[obj.flowsheet().time.first(), ...].fix(
                        0.0
                    )
                except AttributeError:
                    pass

                # Try to fix element_accumulation @ first time point
                try:
                    obj.element_accumulation[obj.flowsheet().time.first(), ...].fix(0.0)
                except AttributeError:
                    pass

                # Try to fix energy_accumulation @ first time point
                try:
                    obj.energy_accumulation[obj.flowsheet().time.first(), ...].fix(0.0)
                except AttributeError:
                    pass

        else:
            raise ValueError(
                "Unrecognised value for argument 'state'. "
                "Valid values are 'steady-state'."
            )

    def unfix_initial_conditions(self):
        """This method unfixed the initial conditions for dynamic models.

        Args:
            None

        Returns :
            None
        """
        for obj in self.component_objects(Block, descend_into=True):
            # Try to unfix material_accumulation @ first time point
            try:
                obj.material_accumulation[obj.flowsheet().time.first(), ...].unfix()
            except AttributeError:
                pass

            # Try to fix element_accumulation @ first time point
            try:
                obj.element_accumulation[obj.flowsheet().time.first(), ...].unfix()
            except AttributeError:
                pass

            # Try to fix energy_accumulation @ first time point
            try:
                obj.energy_accumulation[obj.flowsheet().time.first(), ...].unfix()
            except AttributeError:
                pass

    def report(self, time_point=0, dof=False, ostream=None, prefix=""):

        time_point = float(time_point)

        if ostream is None:
            ostream = sys.stdout

        # Get DoF and model stats
        if dof:
            dof_stat = degrees_of_freedom(self)
            nv = number_variables(self)
            nc = number_activated_constraints(self)
            nb = number_activated_blocks(self)

        # Get components to report in performance section
        performance = self._get_performance_contents(time_point=time_point)

        # Get stream table
        stream_table = self._get_stream_table_contents(time_point=time_point)

        # Set model type output
        if hasattr(self, "is_flowsheet") and self.is_flowsheet:
            model_type = "Flowsheet"
        else:
            model_type = "Unit"

        # Write output
        max_str_length = 84
        tab = " " * 4
        ostream.write("\n" + "=" * max_str_length + "\n")

        lead_str = f"{prefix}{model_type} : {self.name}"
        trail_str = f"Time: {time_point}"
        mid_str = " " * (max_str_length - len(lead_str) - len(trail_str))
        ostream.write(lead_str + mid_str + trail_str)

        if dof:
            ostream.write("\n" + "=" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Local Degrees of Freedom: {dof_stat}")
            ostream.write("\n")
            ostream.write(
                f"{prefix}{tab}Total Variables: {nv}{tab}"
                f"Activated Constraints: {nc}{tab}"
                f"Activated Blocks: {nb}"
            )

        if performance is not None:
            # PYLINT-WHY: pylint has no way of knowing that performance is supposed to be dict-like
            # pylint: disable=unsubscriptable-object
            # PYLINT-TODO: alternatively, have the function return an empty dict and test with `if performance:`
            ostream.write("\n" + "-" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Unit Performance")
            ostream.write("\n" * 2)
            if "vars" in performance.keys() and len(performance["vars"]) > 0:
                ostream.write(f"{prefix}{tab}Variables: \n\n")

                tabular_writer(
                    ostream,
                    prefix + tab,
                    ((k, v) for k, v in performance["vars"].items()),
                    ("Value", "Units", "Fixed", "Bounds"),
                    lambda k, v: [
                        "{:#.5g}".format(report_quantity(v).m),
                        report_quantity(v).u,
                        v.fixed,
                        v.bounds,
                    ],
                )

            if "exprs" in performance.keys() and len(performance["exprs"]) > 0:
                ostream.write("\n")
                ostream.write(f"{prefix}{tab}Expressions: \n\n")

                tabular_writer(
                    ostream,
                    prefix + tab,
                    ((k, v) for k, v in performance["exprs"].items()),
                    (
                        "Value",
                        "Units",
                    ),
                    lambda k, v: [
                        "{:#.5g}".format(report_quantity(v).m),
                        report_quantity(v).u,
                    ],
                )

            if "params" in performance.keys() and len(performance["params"]) > 0:
                ostream.write("\n")
                ostream.write(f"{prefix}{tab}Parameters: \n\n")

                tabular_writer(
                    ostream,
                    prefix + tab,
                    ((k, v) for k, v in performance["params"].items()),
                    ("Value", "Units", "Mutable"),
                    lambda k, v: [
                        report_quantity(v).m,
                        report_quantity(v).u,
                        not v.is_constant(),
                    ],
                )

        if stream_table is not None:
            ostream.write("\n" + "-" * max_str_length + "\n")
            ostream.write(f"{prefix}{tab}Stream Table")
            ostream.write("\n")
            ostream.write(
                textwrap.indent(
                    stream_table_dataframe_to_string(stream_table), prefix + tab
                )
            )
        ostream.write("\n" + "=" * max_str_length + "\n")

    def _get_performance_contents(self, time_point=0):
        return None

    def _get_stream_table_contents(self, time_point=0):
        return None

    def serialize_contents(self, time_point=0):
        """
        Return the performance contents and stream table

        NOTE: There is the possibility of a ConfigurationError because
        the names of the inlets and outlets of the unit model may not be
        standard. If this occurs then return an empty dataframe

        Args:
            time_point: The time

        Returns:
            performance_contents: Pandas dataframe with the performance contents
            stream_table: Pandas dataframe with the stream table for a unit model
        """
        performance_contents = self._get_performance_contents(time_point)
        try:
            stream_table = self._get_stream_table_contents(time_point)
        except ConfigurationError as err:
            _log.warning(f"Could not serialize stream table: {err}")
            stream_table = DataFrame()
        return performance_contents, stream_table

    def _setup_dynamics(self):
        """
        This method automates the setting of the dynamic flag and time domain
        for unit models.

        Performs the following:
         1) Determines if this is a top level flowsheet
         2) Gets dynamic flag from parent if not top level, or checks validity
            of argument provided
         3) Checks has_holdup flag if present and dynamic = True

        Args:
            None

        Returns:
            None
        """
        # Get parent object
        if hasattr(self.parent_block(), "config"):
            # Parent block has a config block, so use this
            parent = self.parent_block()
        else:
            # Use parent flowsheet
            try:
                parent = self.flowsheet()
            except ConfigurationError:
                raise DynamicError(
                    "{} has no parent flowsheet from which to "
                    "get dynamic argument. Please provide a "
                    "value for this argument when constructing "
                    "the unit.".format(self.name)
                )

        # Check the dynamic flag, and retrieve if necessary
        if self.config.dynamic == useDefault:
            # Get flag from parent flowsheet
            try:
                self.config.dynamic = parent.config.dynamic
            except AttributeError:
                # No flowsheet, raise exception
                raise DynamicError(
                    "{} parent flowsheet has no dynamic "
                    "argument. Please provide a "
                    "value for this argument when constructing "
                    "the unit.".format(self.name)
                )

        # Check for case when dynamic=True, but parent dynamic=False
        if self.config.dynamic and not parent.config.dynamic:
            raise DynamicError(
                "{} trying to declare a dynamic model within "
                "a steady-state flowsheet. This is not "
                "supported by the IDAES framework. Try "
                "creating a dynamic flowsheet instead, and "
                "declaring some models as steady-state.".format(self.name)
            )

        # Set and validate has_holdup argument
        if self.config.has_holdup == useDefault:
            # Default to same value as dynamic flag
            self.config.has_holdup = self.config.dynamic
        elif self.config.has_holdup is False:
            if self.config.dynamic is True:
                # Dynamic model must have has_holdup = True
                raise ConfigurationError(
                    "{} invalid arguments for dynamic and has_holdup. "
                    "If dynamic = True, has_holdup must also be True "
                    "(was False)".format(self.name)
                )

    def _get_property_package(self):
        """
        This method gathers the necessary information about the property
        package to be used in the control volume block.

        If a property package has not been provided by the user, the method
        searches up the model tree until it finds an object with the
        'default_property_package' attribute and uses this package for the
        control volume block.

        The method also gathers any default construction arguments specified
        for the property package and combines these with any arguments
        specified by the user for the control volume block (user specified
        arguments take priority over defaults).

        Args:
            None

        Returns:
            None
        """
        # Get property_package block if not provided in arguments
        parent = self.parent_block()
        if self.config.property_package == useDefault:
            # Try to get property_package from parent
            try:
                if parent.config.property_package in [None, useDefault]:
                    parent.config.property_package = self._get_default_prop_pack()

                self.config.property_package = parent.config.property_package
            except AttributeError:
                self.config.property_package = self._get_default_prop_pack()

        # Check for any flowsheet level build arguments
        for k in self.config.property_package.config.default_arguments:
            if k not in self.config.property_package_args:
                self.config.property_package_args[k] = (
                    self.config.property_package.config.default_arguments[k]
                )

    def _get_default_prop_pack(self):
        """
        This method is used to find a default property package defined at the
        flowsheet level if a package is not provided as an argument when
        instantiating the control volume block.

        Args:
            None

        Returns:
            None
        """
        parent = self.flowsheet()
        while True:
            if parent is None:
                raise ConfigurationError(
                    "{} no property package provided and "
                    "no default defined by parent flowsheet(s).".format(self.name)
                )
            elif parent.config.default_property_package is not None:
                _log.info(f"{self.name} Using default property package")
                return parent.config.default_property_package

            parent = parent.flowsheet()

    def _get_indexing_sets(self):
        """
        This method collects all necessary indexing sets from property
        parameter block and makes references to these for use within the
        control volume block. Collected indexing sets are phase_list and
        component_list.

        Args:
            None

        Returns:
            None
        """
        # Check for phase list(s)
        if not hasattr(self.config.property_package, "phase_list"):
            raise PropertyPackageError(
                "{} property_package provided does not "
                "contain a phase_list. "
                "Please contact the developer of the property package.".format(
                    self.name
                )
            )

        # Check for component list(s)
        if not hasattr(self.config.property_package, "component_list"):
            raise PropertyPackageError(
                "{} property_package provided does not "
                "contain a component_list. "
                "Please contact the developer of the property package.".format(
                    self.name
                )
            )

    def _get_reaction_package(self):
        """
        This method gathers the necessary information about the reaction
        package to be used in the control volume block (if required).

        If a reaction package has been provided by the user, the method
        gathers any default construction arguments specified
        for the reaction package and combines these with any arguments
        specified by the user for the control volume block (user specified
        arguments take priority over defaults).

        Args:
            None

        Returns:
            None
        """
        if self.config.reaction_package is not None:
            # Check for any flowsheet level build arguments
            for k in self.config.reaction_package.config.default_arguments:
                if k not in self.config.reaction_package_args:
                    self.config.reaction_package_args[k] = (
                        self.config.reaction_package.config.default_arguments[k]
                    )

    def calculate_scaling_factors(self):
        # This lets you call super().calculate_scaling_factors() in a unit
        # model's calculate_scaling_factors method without worrying about
        # whether the parent class defines one.  This allows for a more standized
        # form of calculate_scaling_factors() methods.
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Generate parameter and expression files for CO2
"""

__author__ = "John Eslick"

import os
import pyomo.environ as pyo
from pyomo.common.fileutils import this_file_dir
from idaes.models.properties.general_helmholtz.helmholtz_parameters import (
    WriteParameters,
)


def thermal_conductivity_rule(m):
    """Thermal conductivity rule

    Fenghour, A., W.A. Wakeham, V. Vesovic, (1998). "The Viscosity of Carbon
        Dioxide." J. Phys. Chem. Ref. Data, 27, 31-44.
    """
    b = {
        0: 0.4226159,
        1: 0.6280115,
        2: -0.5387661,
        3: 0.6735941,
        4: 0,
        5: 0,
        6: -0.4362677,
        7: 0.2255388,
    }
    c = {
        1: 2.387869e-2,
        2: 4.350794,
        3: -10.33404,
        4: 7.981590,
        5: -1.940558,
    }
    d = {
        1: 24.47164,
        2: 8.705605e-2,
        3: -6.547950e-5,
        4: 6.594919e-8,
    }
    T = m.T_star / m.tau
    Ts = T / 251.196
    rho = m.rho_star * m.delta
    G = sum(bval / Ts**i for i, bval in b.items())
    cint_over_k = 1.0 + pyo.exp(-183.5 / T) * sum(
        cval * (T / 100) ** (2 - i) for i, cval in c.items()
    )
    return (
        475.598 * pyo.sqrt(T) * (1 + 2.0 / 5.0 * cint_over_k) / G
        + d[1] * rho
        + d[2] * rho**2
        + d[3] * rho**3
        + d[4] * rho**4
    ) / 1e3


def viscosity_rule(m):
    """Viscosity rule

    Fenghour, A., W.A. Wakeham, V. Vesovic, (1998). "The Viscosity of Carbon
        Dioxide." J. Phys. Chem. Ref. Data, 27, 31-44.
    """
    a = {
        0: 0.235156,
        1: -0.491266,
        2: 5.211155e-2,
        3: 5.347906e-2,
        4: -1.537102e-2,
    }
    d = {
        1: 0.4071119e-2,
        2: 0.7198037e-4,
        3: 0.2411697e-16,
        4: 0.2971072e-22,
        5: -0.1627888e-22,
    }
    T = m.T_star / m.tau
    rho = m.delta * m.rho_star
    Ts = T / 251.196
    return (
        1.00697
        * pyo.sqrt(T)
        / pyo.exp(sum(aval * pyo.log(Ts) ** i for i, aval in a.items()))
        + d[1] * rho
        + d[2] * rho**2
        + d[3] * rho**6 / Ts**3
        + d[4] * rho**8
        + d[5] * rho**8 / Ts
    )


def main(dry_run=False):
    """Generate parameter and expression files.

    Args:
        dry_run (bool): If dry run don't generate files

    Returns:
        None
    """
    main_param_file = os.path.join(this_file_dir(), "co2.json")
    we = WriteParameters(parameters=main_param_file)
    we.add(
        {
            "viscosity": viscosity_rule,
            "thermal_conductivity": thermal_conductivity_rule,
        }
    )
    we.write(dry_run=dry_run)
    return we


if __name__ == "__main__":
    main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

"""
Consider the following optimization problem.

min f:  p1*x1+ p2*(x2^2) + p1*p2
         s.t  c1: x1 + x2 = p1
              c2: x2 + x3 = p2
              0 <= x1, x2, x3 <= 10
              p1 = 10
              p2 = 5

Variables = (x1, x2, x3)
Parameters (fixed variables) = (p1, p2)
"""

import pyomo.environ as pyo
import numpy as np
import pytest
from idaes.apps.uncertainty_propagation.uncertainties import propagate_uncertainty

if __name__ == "__main__":

    ### Create optimization model
    m = pyo.ConcreteModel()
    m.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

    # Define variables
    # m.x1 = pyo.Var(bounds=(0,10))
    # m.x2 = pyo.Var(bounds=(0,10))
    # m.x3 = pyo.Var(bounds=(0,10))

    m.x1 = pyo.Var()
    m.x2 = pyo.Var()
    m.x3 = pyo.Var()

    # Define parameters
    # m.p1 = pyo.Param(initialize=10, mutable=True)
    # m.p2 = pyo.Param(initialize=5, mutable=True)
    # Important Tip: The uncertain parameters need to be defined at Pyomo variables
    m.p1 = pyo.Var(initialize=10)
    m.p2 = pyo.Var(initialize=5)
    # Important Tip: We fix these to first solve with Ipopt. We unfix below
    # before using the uncertainty propagation toolbox.
    m.p1.fix()
    m.p2.fix()

    # Define constraints
    m.con1 = pyo.Constraint(expr=m.x1 + m.x2 - m.p1 == 0)
    m.con2 = pyo.Constraint(expr=m.x2 + m.x3 - m.p2 == 0)

    # Define objective
    m.obj = pyo.Objective(
        expr=m.p1 * m.x1 + m.p2 * (m.x2**2) + m.p1 * m.p2, sense=pyo.minimize
    )

    ### Solve optimization model
    opt = pyo.SolverFactory("ipopt", tee=True)
    opt.solve(m)

    ### Inspect solution
    print("Numeric solution:")
    print("x1 =", m.x1())
    print("x2 =", m.x2())
    print("x3 =", m.x3())
    print(m.dual.display())

    ### Analytic solution
    """
    At the optimal solution, none of the bounds are active. As long as the active set
    does not change (i.e., none of the bounds become active), the
    first order optimality conditions reduce to a simple linear system.
    """

    # dual variables (multipliers)
    v2_ = 0
    v1_ = m.p1()

    # primal variables
    x2_ = (v1_ + v2_) / (2 * m.p2())
    x1_ = m.p1() - x2_
    x3_ = m.p2() - x2_

    print("\nAnalytic solution:")
    print("x1 =", x1_)
    print("x2 =", x2_)
    print("x3 =", x3_)
    print("v1 =", v1_)
    print("v2 =", v2_)

    ### Analytic sensitivity
    """
    Using the analytic solution above, we can compute the sensitivies of x and v to
    perturbations in p1 and p2.


    The matrix dx_dp constrains the sensitivities of x to perturbations in p
    """

    # Initialize sensitivity matrix Nx x Np
    # Rows: variables x
    # Columns: parameters p
    dx_dp = np.zeros((3, 2))

    # dx2/dp1 = 1/(2 * p2)
    dx_dp[1, 0] = 1 / (2 * m.p2())

    # dx2/dp2 = -(v1 + v2)/(2 * p2**2)
    dx_dp[1, 1] = -(v1_ + v2_) / (2 * m.p2() ** 2)

    # dx1/dp1 = 1 - dx2/dp1
    dx_dp[0, 0] = 1 - dx_dp[1, 0]

    # dx1/dp2 = 0 - dx2/dp2
    dx_dp[0, 1] = 0 - dx_dp[1, 1]

    # dx3/dp1 = 1 - dx2/dp1
    dx_dp[2, 0] = 0 - dx_dp[1, 0]

    # dx3/dp2 = 0 - dx2/dp2
    dx_dp[2, 1] = 1 - dx_dp[1, 1]

    print("\n\ndx/dp =\n", dx_dp)

    """
    Similarly, we can compute the gradients df_dx, df_dp
    and Jacobians dc_dx, dc_dp
    """

    # Initialize 1 x 3 array to store (\partial f)/(\partial x)
    # Elements: variables x
    df_dx = np.zeros(3)

    # df/dx1 = p1
    df_dx[0] = m.p1()

    # df/dx2 = p2
    df_dx[1] = 2 * m.p2() * x2_

    # df/dx3 = 0

    print("\n\ndf/dx =\n", df_dx)

    # Initialize 1 x 2 array to store (\partial f)/(\partial p)
    # Elements: parameters p
    df_dp = np.zeros(2)

    # df/dxp1 = x1 + p2
    df_dp[0] = x1_ + m.p2()

    # df/dp2 = x2**2 + p1
    df_dp[1] = x2_**2 + m.p1()

    print("\n\ndf/dp =\n", df_dp)

    # Initialize 2 x 3 array to store (\partial c)/(\partial x)
    # Rows: constraints c
    # Columns: variables x
    dc_dx = np.zeros((2, 3))

    # dc1/dx1 = 1
    dc_dx[0, 0] = 1

    # dc1/dx2 = 1
    dc_dx[0, 1] = 1

    # dc2/dx2 = 1
    dc_dx[1, 1] = 1

    # dc2/dx3 = 1
    dc_dx[1, 2] = 1

    # Remaining entries are 0

    print("\n\ndc/dx =\n", dc_dx)

    # Initialize 2 x 2 array to store (\partial c)/(\partial x)
    # Rows: constraints c
    # Columns: variables x
    dc_dp = np.zeros((2, 2))

    # dc1/dp1 = -1
    dc_dp[0, 0] = -1

    # dc2/dp2 = -1
    dc_dp[1, 1] = -1

    # Remaining entries are 0

    print("\n\ndc/dp =\n", dc_dp)

    ### Uncertainty propagation
    """
    Now lets test the uncertainty propagation package. We will assume p has covariance
    sigma_p = [[2, 0], [0, 1]]
    """

    ## Prepare inputs

    # Covariance matrix
    sigma_p = np.array([[2, 0], [0, 1]])

    # Nominal values for uncertain parameters
    theta = {"p1": m.p1(), "p2": m.p2()}

    # Names of uncertain parameters
    theta_names = ["p1", "p2"]

    # Important to unfix the parameters!
    # Otherwise k_aug will complain about too few degrees of freedom
    m.p1.unfix()
    m.p2.unfix()

    ## Run sensitivity toolbox
    results = propagate_uncertainty(m, theta, sigma_p, theta_names)

    ## Compute uncertainty propagation by hand
    # We now compute the expected result with our
    # analytic solution.

    # Take 1: use the formula in the documentation
    tmp_f = df_dp + df_dx @ dx_dp
    sigma_f = tmp_f @ sigma_p @ tmp_f.transpose()
    print("\nsigma_f = ", sigma_f)

    tmp_c = dc_dp + dc_dx @ dx_dp
    sigma_c = tmp_c @ sigma_p @ tmp_c.transpose()
    print("\nsigma_c = \n", sigma_c)

    # Take 2: use an equivalent formula
    print("\n\n")
    grad_f = np.hstack((df_dx, df_dp))
    dsdp = np.vstack((dx_dp, np.eye(2)))
    sigma_f_take_2 = grad_f @ dsdp @ sigma_p @ dsdp.transpose() @ grad_f.transpose()
    # This matches
    print("sigma_f (take 2) = ", sigma_f_take_2)

    dc_ds = np.hstack((dc_dx, dc_dp))
    sigma_c_take_2 = dc_ds @ dsdp @ sigma_p @ dsdp.transpose() @ dc_ds.transpose()
    # This matches
    print("sigma_c (take 2) = ", sigma_c_take_2)

    ## Check results
    # We now check the results for the uncertainty propagation toolbox

    # Check the order from the uncertainty propagation toolbox is as expected
    # TODO: make this test more automated to use this information
    # to reshuffle the analytic solution
    assert results.col == ["x1", "x2", "p1", "p2", "x3"]
    assert results.row == ["con1", "con2", "obj"]

    # Define indices to compare results to analytic solution
    # TODO: make this more robust to use the results above
    var_idx = np.array([True, True, False, False, True])
    theta_idx = np.array([False, False, True, True, False])

    # Check the gradient of the objective w.r.t. x matches
    np.testing.assert_array_almost_equal(results.gradient_f[var_idx], np.array(df_dx))

    # Check the gradient of the objective w.r.t. p (parameters) matches
    np.testing.assert_array_almost_equal(results.gradient_f[theta_idx], np.array(df_dp))

    # Check the Jacobian of the constraints w.r.t. x matches
    np.testing.assert_array_almost_equal(
        results.gradient_c.toarray()[:, var_idx], np.array(dc_dx)
    )

    # Check the Jacobian of the constraints w.r.t. p (parameters) matches
    np.testing.assert_array_almost_equal(
        results.gradient_c.toarray()[:, theta_idx], np.array(dc_dp)
    )

    # Check the NLP sensitivity results for the variables (x) matches
    np.testing.assert_array_almost_equal(
        results.dsdp.toarray()[var_idx, :], np.array(dx_dp)
    )

    # Check the NLP sensitivity results for the parameters (p) matches
    np.testing.assert_array_almost_equal(
        results.dsdp.toarray()[theta_idx, :], np.array([[1, 0], [0, 1]])
    )

    # Check the uncertainty propagation results for the constrains matches
    np.testing.assert_array_almost_equal(results.propagation_c, np.sum(sigma_c))

    # Check the uncertainty propagation results for the objective matches
    assert results.propagation_f == pytest.approx(sigma_f)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test for Caprese's module for NMPC.
"""

import pytest
from pyomo.environ import Var
from pyomo.common.collections import ComponentSet
from pyomo.dae.flatten import flatten_dae_components

from idaes.apps.caprese.categorize import (
    categorize_dae_variables,
)
from idaes.apps.caprese.common.config import VariableCategory as VC
from idaes.apps.caprese.examples.cstr_model import make_model

__author__ = "Robert Parker"


@pytest.mark.unit
def test_categorize_1():
    """
    This test categorizes the enzyme cstr "as intended."
    Volume is used as a measurement, and solvent flow rates are
    fixed, but otherwise equations are as expected.
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    init_input_list = [
        model.fs.mixer.S_inlet.flow_vol[0],
        model.fs.mixer.E_inlet.flow_vol[0],
    ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
        model.fs.cstr.control_volume.energy_accumulation[0, "aq"],
        *list(model.fs.cstr.control_volume.material_accumulation[0, "aq", :]),
    ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
        model.fs.cstr.control_volume.energy_holdup[0, "aq"],
        *list(model.fs.cstr.control_volume.material_holdup[0, "aq", :]),
    ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
        model.fs.mixer.E_inlet.temperature[0],
        model.fs.mixer.S_inlet.temperature[0],
        *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
        *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
        model.fs.cstr.outlet.conc_mol[0, "Solvent"],
        model.fs.mixer.outlet.conc_mol[0, "Solvent"],
    ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
        model.fs.cstr.control_volume.energy_holdup[0, "aq"],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.material_holdup[0, "aq", :]),
    ]
    init_meas_set = ComponentSet(init_meas_list)
    # Solvent holdup is not a measurement; we measure volume instead
    init_meas_set.remove(
        model.fs.cstr.control_volume.material_holdup[0, "aq", "Solvent"]
    )

    init_alg_list = [
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, "aq", :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
    ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, "Solvent"])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, "Solvent"])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_2():
    """
    In this instance, temperature is "measured" (used as an initial condition)
    instead of energy_holdup, conc[P] is measured instead of holdup[P],
    and accumulation[C] is measured instead of holdup[C].
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time
    t0 = time.first()

    material_holdup = model.fs.cstr.control_volume.material_holdup
    material_accumulation = model.fs.cstr.control_volume.material_accumulation
    energy_holdup = model.fs.cstr.control_volume.energy_holdup
    energy_accumulation = model.fs.cstr.control_volume.energy_accumulation
    conc_mol = model.fs.cstr.outlet.conc_mol

    # Specify temperature instead of energy holdup
    energy_holdup[t0, "aq"].unfix()
    model.fs.cstr.outlet.temperature[t0].fix(300)

    # Specify C_P instead of holdup
    material_holdup[t0, "aq", "P"].unfix()
    conc_mol[t0, "P"].fix(0.0)

    # Specify accumulation of C instead of holdup
    # You might want to do this if you want to start at steady state,
    # but don't know the value of every variable at the steady state
    # you want to start at...
    material_holdup[t0, "aq", "C"].unfix()
    material_accumulation[t0, "aq", "C"].fix(0.0)

    init_input_list = [
        model.fs.mixer.S_inlet.flow_vol[0],
        model.fs.mixer.E_inlet.flow_vol[0],
    ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
        model.fs.cstr.control_volume.energy_accumulation[0, "aq"],
        *list(model.fs.cstr.control_volume.material_accumulation[0, "aq", :]),
    ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
        model.fs.cstr.control_volume.energy_holdup[0, "aq"],
        *list(model.fs.cstr.control_volume.material_holdup[0, "aq", :]),
    ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
        model.fs.mixer.E_inlet.temperature[0],
        model.fs.mixer.S_inlet.temperature[0],
        *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
        *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
        model.fs.cstr.outlet.conc_mol[0, "Solvent"],
        model.fs.mixer.outlet.conc_mol[0, "Solvent"],
    ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        model.fs.cstr.control_volume.material_holdup[0, "aq", "S"],
        model.fs.cstr.control_volume.material_holdup[0, "aq", "E"],
        model.fs.cstr.control_volume.material_holdup[0, "aq", "S"],
        model.fs.cstr.control_volume.material_accumulation[0, "aq", "C"],
        model.fs.cstr.outlet.conc_mol[0, "P"],
    ]
    init_meas_set = ComponentSet(init_meas_list)
    # No need to remove solvent holdup here as it was not added to this list.

    init_alg_list = [
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, "aq", :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
    ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, "Solvent"])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, "Solvent"])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_3():
    """
    In this test we fix one of the differential variables.
    This is the case where somebody wants to run an isothermal
    CSTR.
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    CV = model.fs.cstr.control_volume
    CV.energy_holdup.fix(300)
    CV.energy_accumulation_disc_eq.deactivate()

    init_input_list = [
        model.fs.mixer.S_inlet.flow_vol[0],
        model.fs.mixer.E_inlet.flow_vol[0],
    ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
        *list(model.fs.cstr.control_volume.material_accumulation[0, "aq", :]),
    ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
        *list(model.fs.cstr.control_volume.material_holdup[0, "aq", :]),
    ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
        # Energy holdup has been fixed
        model.fs.cstr.control_volume.energy_holdup[0, "aq"],
        model.fs.mixer.E_inlet.temperature[0],
        model.fs.mixer.S_inlet.temperature[0],
        *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
        *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
        model.fs.cstr.outlet.conc_mol[0, "Solvent"],
        model.fs.mixer.outlet.conc_mol[0, "Solvent"],
    ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.material_holdup[0, "aq", :]),
    ]
    init_meas_set = ComponentSet(init_meas_list)
    # Solvent holdup is not a measurement; we measure volume instead
    init_meas_set.remove(
        model.fs.cstr.control_volume.material_holdup[0, "aq", "Solvent"]
    )

    init_alg_list = [
        # Since energy_holdup is fixed, energy_accumulation is not
        # considered a derivative. Instead it is algebraic.
        model.fs.cstr.control_volume.energy_accumulation[0, "aq"],
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, "aq", :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
    ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, "Solvent"])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, "Solvent"])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_4():
    """
    This tests categorization when a pseudo-steady state
    approximation is used. Energy accumulation and accumulation of E
    are fixed, the corresponding initial conditions unfixed, and
    the corresponding discretization equations deactivated.
    """
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    CV = model.fs.cstr.control_volume
    CV.energy_accumulation[:, "aq"].fix(0.0)
    CV.material_accumulation[:, "aq", "E"].fix(0.0)
    CV.energy_holdup[0, "aq"].unfix()
    CV.material_holdup[0, "aq", "E"].unfix()
    CV.energy_accumulation_disc_eq.deactivate()
    CV.material_accumulation_disc_eq.deactivate()

    init_input_list = [
        model.fs.mixer.S_inlet.flow_vol[0],
        model.fs.mixer.E_inlet.flow_vol[0],
    ]
    init_input_set = ComponentSet(init_input_list)

    init_deriv_list = [
        CV.material_accumulation[0, "aq", "C"],
        CV.material_accumulation[0, "aq", "S"],
        CV.material_accumulation[0, "aq", "P"],
        CV.material_accumulation[0, "aq", "Solvent"],
    ]
    init_deriv_set = ComponentSet(init_deriv_list)

    init_diff_list = [
        CV.material_holdup[0, "aq", "C"],
        CV.material_holdup[0, "aq", "S"],
        CV.material_holdup[0, "aq", "P"],
        CV.material_holdup[0, "aq", "Solvent"],
    ]
    init_diff_set = ComponentSet(init_diff_list)

    init_fixed_list = [
        model.fs.cstr.control_volume.energy_accumulation[0, "aq"],
        CV.material_accumulation[0, "aq", "E"],
        model.fs.mixer.E_inlet.temperature[0],
        model.fs.mixer.S_inlet.temperature[0],
        *list(model.fs.mixer.E_inlet.conc_mol[0, :]),
        *list(model.fs.mixer.S_inlet.conc_mol[0, :]),
        model.fs.cstr.outlet.conc_mol[0, "Solvent"],
        model.fs.mixer.outlet.conc_mol[0, "Solvent"],
    ]
    init_fixed_set = ComponentSet(init_fixed_list)

    init_meas_list = [
        model.fs.cstr.control_volume.volume[0],
        CV.material_holdup[0, "aq", "P"],
        CV.material_holdup[0, "aq", "C"],
        CV.material_holdup[0, "aq", "S"],
    ]
    init_meas_set = ComponentSet(init_meas_list)

    init_alg_list = [
        model.fs.cstr.control_volume.energy_holdup[0, "aq"],
        CV.material_holdup[0, "aq", "E"],
        model.fs.cstr.outlet.flow_vol[0],
        model.fs.cstr.outlet.temperature[0],
        model.fs.cstr.inlet.flow_vol[0],
        model.fs.cstr.inlet.temperature[0],
        model.fs.mixer.outlet.flow_vol[0],
        model.fs.mixer.outlet.temperature[0],
        model.fs.cstr.control_volume.volume[0],
        *list(model.fs.cstr.control_volume.properties_out[0].flow_mol_comp[:]),
        *list(model.fs.cstr.inlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.properties_in[0].flow_mol_comp[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_generation[0, "aq", :]),
        *list(model.fs.mixer.mixed_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.E_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.mixer.S_inlet_state[0].flow_mol_comp[:]),
        *list(model.fs.cstr.outlet.conc_mol[0, :]),
        *list(model.fs.mixer.outlet.conc_mol[0, :]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_coef[:]),
        *list(model.fs.cstr.control_volume.reactions[0].reaction_rate[:]),
        *list(model.fs.cstr.control_volume.rate_reaction_extent[0, :]),
    ]
    init_alg_set = ComponentSet(init_alg_list)
    # solvent outlet concentrations are not algebraic variables as
    # it is fixed.
    init_alg_set.remove(model.fs.cstr.outlet.conc_mol[0, "Solvent"])
    init_alg_set.remove(model.fs.mixer.outlet.conc_mol[0, "Solvent"])

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(dae_vars, time, init_input_list)

    input_vars = category_dict[VC.INPUT]
    diff_vars = category_dict[VC.DIFFERENTIAL]
    deriv_vars = category_dict[VC.DERIVATIVE]
    fixed_vars = category_dict[VC.FIXED]
    alg_vars = category_dict[VC.ALGEBRAIC]
    meas_vars = category_dict[VC.MEASUREMENT]

    assert len(input_vars) == len(init_input_set)
    for v in input_vars:
        assert v[0] in init_input_set

    assert len(deriv_vars) == len(init_deriv_set)
    for v in deriv_vars:
        assert v[0] in init_deriv_set

    assert len(diff_vars) == len(init_deriv_set)
    for v in diff_vars:
        assert v[0] in init_diff_set

    assert len(fixed_vars) == len(init_fixed_set)
    for v in fixed_vars:
        assert v[0] in init_fixed_set

    assert len(alg_vars) == len(init_alg_set)
    for v in alg_vars:
        assert v[0] in init_alg_set

    assert len(meas_vars) == len(init_meas_set)
    for v in meas_vars:
        assert v[0] in init_meas_set

    assert len(scalar_vars) == 0


@pytest.mark.unit
def test_categorize_error():
    model = make_model(horizon=1, ntfe=5, ntcp=2)
    time = model.fs.time

    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)

    # Add a dummy var to treat as an input.
    # This var is not in `dae_vars`, so it will not be located during
    # categorization, which should fail.
    model.dummy_var = Var(time)

    init_input_list = [
        model.fs.mixer.S_inlet.flow_vol[0],
        model.fs.mixer.E_inlet.flow_vol[0],
        model.dummy_var[0],
    ]

    with pytest.raises(RuntimeError, match=r"Not all inputs could be found"):
        category_dict = categorize_dae_variables(
            dae_vars,
            time,
            init_input_list,
        )

    # Re-run flattener. Now `dummy_var` should be included in `dae_vars`.
    scalar_vars, dae_vars = flatten_dae_components(model, time, ctype=Var)
    category_dict = categorize_dae_variables(
        dae_vars,
        time,
        init_input_list,
    )


if __name__ == "__main__":
    test_categorize_1()
    test_categorize_2()
    test_categorize_3()
    test_categorize_4()
    test_categorize_error()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""This module contains utility classes that allow users to tag model quantities
and group them, for easy display, formatting, and input.
"""
import xml.dom.minidom
import collections

import pyomo.environ as pyo
from pyomo.core.base.indexed_component_slice import IndexedComponent_slice

import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)

__Author__ = "John Eslick"


class ModelTag:
    """The purpose of this class is to facilitate a simpler method of accessing,
    displaying, and reporting model quantities. The structure of IDAES models is
    a complex hierarchy. This class allows quantities to be accessed more directly
    and provides more control over how they are reported."""

    __slots__ = [
        "_format",
        "_expression",
        "_doc",
        "_display_units",
        "_cache_validation_value",
        "_cache_display_value",
        "_name",
        "_root",
        "_index",
        "_group",
        "_str_units",
        "_set_in_display_units",
    ]

    def __init__(self, expr, format_string="{}", doc="", display_units=None):
        """initialize a model tag instance.
        Args:
            expr: A Pyomo Var, Expression, Param, Reference, or unnamed
                expression to tag. This can be a scalar or indexed.
            format_string: A formatting string used to print an element of the
                tagged expression (e.g. '{:.3f}').
            doc: A description of the tagged quantity.
            display_units: Pyomo units to display the quantity in. If a string
                is provided it will be used to display as the unit, but will not
                be used to convert units. If None, use native units of the
                quantity.
        """
        super().__init__()
        self._format = format_string  # format string for printing expression
        if isinstance(expr, IndexedComponent_slice):
            self._expression = pyo.Reference(expr)  # tag expression (can be unnamed)
        else:
            self._expression = expr
        self._doc = doc  # documentation for a tag
        self._display_units = display_units  # unit to display value in
        self._cache_validation_value = {}  # value when converted value stored
        self._cache_display_value = {}  # value to display after unit conversions
        self._name = None  # tag name (just used to clarify error messages)
        self._root = None  # use this to cache scalar tags in indexed parent
        self._index = None  # index to get cached converted value from parent
        self._group = None  # Group object if this is a member of a group
        self._str_units = True  # include units to stringify the tag
        # if _set_in_display_units is True and no units are provided for for
        # set, fix, setub, and setlb, the units will be assumed to be the
        # display units.  If it is false and no units are provided, the units are
        # assumed to be the native units of the quantity
        self._set_in_display_units = False

    def __getitem__(self, k):
        """Returns a new ModelTag for a scalar element of a tagged indexed
        quantity or a ModelTag with a slice as the expression."""
        if k is None and not self.is_indexed:
            return self
        try:
            tag = ModelTag(
                expr=self.expression[k],
                format_string=self._format,
                display_units=self._display_units,
                doc=self._doc,
            )
        except KeyError as key_err:
            if self._name is None:
                raise KeyError(f"{k} is not a valid index for tag") from key_err
            raise KeyError(
                f"{k} is not a valid index for tag {self._name}"
            ) from key_err
        if self._root is None:  # cache the unit conversion in root object
            tag._root = self
        else:
            tag._root = self._root
        tag._index = k
        return tag

    def __len__(self):
        """Number of elements in a tag"""
        return len(self.indexes)

    def keys(self):
        """Iterator of keys in the tag.  If scalar, is a single None."""
        indx_set = self.indexes
        for i in indx_set:
            yield i

    def values(self):
        """Iterator for scalar elements in a tag (The elements are scalar tags)"""
        indx_set = self.indexes
        for i in indx_set:
            yield self[i]

    def items(self):
        """Iterator for key scalar elements pairs in a tag"""
        indx_set = self.indexes
        for i in indx_set:
            yield i, self[i]

    def __str__(self):
        """Returns the default string representation of a tagged quantity. If
        the tagged expression is indexed this uses the default index.  This can
        be handy for things like the time index."""
        return self.display(units=self.str_include_units)

    def __call__(self, *args, **kwargs):
        """Calling an instance of a tagged quantity gets the display string see
        display()"""
        return self.display(*args, **kwargs)

    def display(self, units=True, format_string=None, index=None):
        """Get a string representation of the tagged quantity

        Args:
            units (bool): Include units of measure in the string
            format_string (str): Formatting string, if supplied overrides the default
            index: If the tagged quantity is indexed, an index for the element
                to display is required, or the default index is used.

        Returns:
            str
        """
        val = self.get_display_value(index=index)
        if format_string is None:
            format_string = self.get_format(units=False, index=index)
        if callable(format_string):
            format_string = format_string(val)
        if units:
            format_string = self._join_units(index=index, format_string=format_string)
        try:
            return format_string.format(val)
        except ValueError:
            # Probably trying to put a string through the number format
            # for various reasons, I'll allow it.
            return str(val)
        except TypeError:
            # Probably trying to put None through the numeric format.  This
            # can happen for example when variables don't have values.  I'll
            # allow 'None' to be printed.  It's not uncommon to happen, and I
            # don't want to raise an exception.
            return str(val)

    def _join_units(self, index=None, format_string=None):
        """Private method to join the format string with the units of measure
        string.
        """
        if format_string is None:
            format_string = self._format
        units = self.get_unit_str(index=index)
        if units in ["None", "", "dimensionless"] or units is None:
            return format_string
        if units == "%":
            return "".join([format_string, units])
        return " ".join([format_string, units])

    @property
    def expression(self):
        """The tagged expression"""
        return self._expression

    @property
    def value(self):
        """The numeric value of the tagged expression in the display units"""
        if not self.is_indexed:
            return self.get_display_value()
        else:
            val_dict = dict.fromkeys(self.indexes)
            for k, v in self.items():
                val_dict[k] = v.get_display_value()
            return val_dict

    @property
    def native_value(self):
        """The numeric value of the tagged expression in the native units"""
        if not self.is_indexed:
            return self.get_display_value(convert=False)
        else:
            val_dict = dict.fromkeys(self.indexes)
            for k, v in self.items():
                val_dict[k] = v.get_display_value(convert=False)
            return val_dict

    @property
    def doc(self):
        """Tag documentation string"""
        return self._doc

    def get_format(self, units=True, index=None):
        """Get the formatting string.

        Args:
            units: if True include units of measure
            index: index for indexed expressions

        Returns
            str:
        """
        if units:
            return self._join_units(index=index)
        return self._format

    def get_display_value(self, index=None, convert=True):
        """Get the value of the expression to display.  Do unit conversion if
        needed.  This caches the unit conversion, to save time if this is called
        repeatedly.  The unconverted value is used to ensure the cached
        converted value is still valid.

        Args:
            index: index of value to display if expression is indexed
            convert: if False don't do unit conversion

        Returns:
            numeric expression value
        """
        if self._root is not None:
            if not self.is_indexed:
                index = self._index
            return self._root.get_display_value(index=index, convert=convert)

        try:
            expr = self.expression[index]
        except KeyError as key_err:
            if self._name is None:
                raise KeyError(f"{index} not a valid key for tag") from key_err
            raise KeyError(f"{index} not a valid key for tag {self._name}") from key_err
        except TypeError:
            expr = self.expression

        if (
            self._display_units is None
            or isinstance(self._display_units, str)
            or not convert
        ):
            # no display units, so display in original units, no convert opt
            try:
                return pyo.value(expr, exception=False)
            except ZeroDivisionError:
                return "ZeroDivisionError"
            except ValueError:  # no value
                return None

        try:
            val = pyo.value(expr, exception=False)
        except ZeroDivisionError:
            return "ZeroDivisionError"
        except ValueError:  # no value
            return None

        cache_validate = self._cache_validation_value
        cache_value = self._cache_display_value
        if index in cache_validate and val == cache_validate[index]:
            return cache_value[index]
        cache_validate[index] = val
        if self._display_units is None:
            cache_value[index] = pyo.value(expr, exception=False)
        else:
            try:
                cache_value[index] = pyo.value(
                    pyo.units.convert(expr, self._display_units), exception=False
                )
            except ValueError:
                return None
        return cache_value[index]

    def get_unit_str(self, index=None):
        """String representation of the tagged quantity's units of measure"""
        if self._display_units is None:
            if self.is_indexed:
                return str(pyo.units.get_units(self.expression[index]))
            return str(pyo.units.get_units(self.expression))
        return str(self._display_units)

    @property
    def is_var(self):
        """Whether the tagged expression is a Pyomo Var. Tagged variables
        can be fixed or set, while expressions cannot.
        """
        try:
            return issubclass(self._expression.ctype, pyo.Var)
        except AttributeError:
            return False

    @property
    def fixed(self):
        """Get the tagged variable if the tag is not a variable, raise TypeError"""
        if not self.is_var:
            return False
        return self.expression.fixed

    @property
    def is_indexed(self):
        """Returns whether the tagged expression is an indexed."""
        try:
            return self.expression.is_indexed()
        except AttributeError:
            return False

    @property
    def indexes(self):
        """The index set of the tagged quantity"""
        if self.is_indexed:
            return list(self.expression.keys())
        return (None,)

    @property
    def group(self):
        """The ModelTagGroup object that this belongs to, if any."""
        if self._root is not None:
            return self._root.group
        return self._group

    @group.setter
    def group(self, val):
        """The ModelTagGroup object that this belongs to, if any."""
        if self._root is not None:
            raise RuntimeError("group is superseded by the root property.")
        self._group = val

    @property
    def str_include_units(self):
        """Set whether to include units by default in the tag's string
        representation"""
        if self.group is not None:
            return self.group.str_include_units
        if self._root is not None:
            return self._root.str_include_units
        return self._str_units

    @str_include_units.setter
    def str_include_units(self, val):
        if self._group is not None:
            raise RuntimeError("str_include_units is superseded by the group property.")
        if self._root is not None:
            raise RuntimeError("str_include_units is superseded by the root property.")
        self._str_units = val

    @property
    def set_in_display_units(self):
        """Default index to use in the tag's string representation, this
        is required for indexed quantities if you want to automatically convert
        to string. An example use it for a time indexed tag, to display a
        specific time point."""
        if self.group is not None:
            return self.group.set_in_display_units
        if self._root is not None:
            return self._root.set_in_display_units
        return self._set_in_display_units

    @set_in_display_units.setter
    def set_in_display_units(self, val):
        """Default index to use in the tag's string representation, this
        is required for indexed quantities if you want to automatically convert
        to string. An example use it for a time indexed tag, to display a
        specific time point."""
        if self.group is not None:
            raise RuntimeError(
                "set_in_display_units is superseded by the group property."
            )
        if self._root is not None:
            raise RuntimeError(
                "set_in_display_units is superseded by the root property."
            )
        self._set_in_display_units = val

    def set(self, val, in_display_units=None):
        """Set the value of a tagged variable.

        Args:
            val: value
            in_display_units: if true assume the value is in the display units

        Returns:
            None
        """
        # If tag is a common immutable then allow it to be set
        if self.expression is None or isinstance(
            self.expression, (str, float, int, tuple)
        ):
            self._expression = val
            return

        if in_display_units is None:
            in_display_units = self.set_in_display_units

        vu = pyo.units.get_units(val)
        if in_display_units and (vu is None or vu == pyo.units.dimensionless):
            if self._display_units is not None:
                val *= self._display_units
        try:
            try:
                self.expression.set_value(val)
            except ValueError:  # it's an indexed expression or slice
                for index in self.expression:
                    self.expression[index].set_value(val)
        except AttributeError as attr_err:
            if self._name:
                raise AttributeError(
                    f"Tagged expression {self._name}, has no set_value()."
                ) from attr_err
            raise AttributeError("Tagged expression has no set_value().") from attr_err

    def setlb(self, val, in_display_units=None):
        """Set the lower bound of a tagged variable.

        Args:
            v: value
            in_display_units: if true assume the value is in the display units

        Returns:
            None
        """
        if in_display_units is None:
            in_display_units = self.set_in_display_units

        vu = pyo.units.get_units(val)
        if in_display_units and (vu is None or vu == pyo.units.dimensionless):
            if self._display_units is not None:
                val *= self._display_units

        try:
            try:
                self.expression.setlb(val)
            except ValueError:  # it's an indexed expression or slice
                for index in self.expression:
                    self.expression[index].lb(val)
        except AttributeError as attr_err:
            if self._name:
                raise AttributeError(
                    f"Tagged expression {self._name}, has no setlb()."
                ) from attr_err
            raise AttributeError("Tagged expression has no setlb().") from attr_err

    def setub(self, val, in_display_units=None):
        """Set the value of a tagged variable.

        Args:
            v: value
            in_display_units: if true assume the value is in the display units

        Returns:
            None
        """
        if in_display_units is None:
            in_display_units = self.set_in_display_units

        vu = pyo.units.get_units(val)
        if in_display_units and (vu is None or vu == pyo.units.dimensionless):
            if self._display_units is not None:
                val *= self._display_units

        try:
            try:
                self.expression.setub(val)
            except ValueError:  # it's an indexed expression or slice
                for index in self.expression:
                    self.expression[index].setub(val)
        except AttributeError as attr_err:
            if self._name:
                raise AttributeError(
                    f"Tagged expression {self._name}, has no setub()."
                ) from attr_err
            raise AttributeError("Tagged expression has no setub().") from attr_err

    def fix(self, val=None, in_display_units=None):
        """Fix the value of a tagged variable.

        Args:
            val: value, if None fix without setting a value
            in_display_units: if true assume the value is in the display units

        Returns:
            None
        """
        if in_display_units is None:
            in_display_units = self.set_in_display_units

        vu = pyo.units.get_units(val)
        if in_display_units and (vu is None or vu == pyo.units.dimensionless):
            if self._display_units is not None:
                val *= self._display_units

        try:
            if val is None:
                self.expression.fix()
            else:
                self.expression.fix(val)
        except AttributeError as attr_err:
            if self._name:
                raise AttributeError(f"Tag {self._name} has no fix.") from attr_err
            raise AttributeError("Tag has no fix.") from attr_err

    def unfix(self):
        """Unfix the value of a tagged variable.

        Args:
            v: value, if None fix without setting a value

        Returns:
            None
        """
        try:
            self.expression.unfix()
        except AttributeError as attr_err:
            if self._name:
                raise AttributeError(f"Tag, {self._name}, has no unfix.") from attr_err
            raise AttributeError("Tag has no unfix.") from attr_err

    @property
    def var(self):
        """Get the tagged variable if the tag is not a variable, raise TypeError"""
        if not self.is_var:
            raise TypeError(
                "Can only return a variable if the expression is a variable."
            )
        return self.expression


class ModelTagGroup(dict):
    """This a basically a dictionary of ModelTag objects. This is used to group
    sets of tags, and contains methods to operate on sets of tags. It inherits dict
    so dictionary methods can be used."""

    __slots__ = [
        "_str_units",
        "_set_in_display_units",
    ]

    def __init__(self):
        super().__init__()
        self._str_units = True
        self._set_in_display_units = False

    def __setitem__(self, key, val):
        if isinstance(val, ModelTag):
            super().__setitem__(key, val)
            self[key]._name = key
            self[key]._group = self
        else:
            raise TypeError("Only ModelTag objects can be part of a ModelTagGroup.")

    def add(self, name, expr, **kwargs):
        """Add a new model tag to the group"""
        if isinstance(expr, ModelTag):
            self[name] = expr
        else:
            self[name] = ModelTag(expr=expr, **kwargs)

    def _table_tagkey_index_lists(self, tags=None):
        """PRIVATE utility function to take tags (list of tag keys or 2-element
        list of a tag key and index) and turn it into a list of tag keys and list
        of indexes. Indexed tags with no index will be expanded to include all
        indexes.
        """

        if tags is None:
            tags = list(self.keys())

        tag_list = []
        indexes = []
        for i, tag in enumerate(tags):  # pylint: disable=unused-variable
            if not isinstance(tag, collections.abc.Hashable):
                if len(tag) == 2:
                    tag_list.append(tag[0])
                    indexes.append(tag[1])
                else:
                    raise ValueError("Key-index pairs should be a list of length 2")
            else:
                if not self[tag].is_indexed:
                    tag_list.append(tag)
                    indexes.append(None)
                else:
                    for k in self[tag].keys():
                        tag_list.append(tag)
                        indexes.append(k)
        return tag_list, indexes

    def table_heading(self, tags=None, units=True):
        """Create a table heading with a given set of tags, for tabulating model
        results.

        Args:
            tags: List (not tuple, since a tuple can be a key) of tag keys or
                2-element list of tags key and an index. If a key is for an
                indexed value and no index is given with the key, it will be
                flattened to create a column for each index.
            units: If True, include the units of measure in the column heading
                names.

        Returns:
            list of column headings
        """
        tag_list, indexes = self._table_tagkey_index_lists(tags)

        row = [None] * len(tag_list)
        for i, tag in enumerate(tag_list):
            if indexes[i] is None:
                name = tag
            else:
                name = f"{tag}[{indexes[i]}]"
            if units:
                row[i] = f"{name} ({self[tag].get_unit_str(index=indexes[i])})"
            else:
                row[i] = name
        return row

    def table_row(self, tags=None, units=True, numeric=False):
        """Create a table row with a given set of tags, for tabulating model
        results.  The row contains values of tagged quantities either as strings
        of numeric values.

        Args:
            tags: List (not tuple, since a tuple can be a key) of tag keys or
                2-element list of tags key and an index. If a key is for an
                indexed value and no index is given with the key, it will be
                flattened to create a column for each index.
            units: If true include the units of measure in the value, if the
                values are not numeric.
            numeric: If true provide numeric values rather than a formatted
                string.

        Returns:
            list of values for a table row either numeric or as formatted strings
        """
        tag_list, indexes = self._table_tagkey_index_lists(tags)

        row = [None] * len(tag_list)
        for i, tag in enumerate(tag_list):
            if numeric:
                row[i] = self[tag][indexes[i]].value
            else:
                row[i] = self[tag][indexes[i]].display(units=units)
        return row

    @property
    def str_include_units(self):
        """When converting a tag in this group directly to a string, include units
        or not.
        """
        return self._str_units

    @str_include_units.setter
    def str_include_units(self, value):
        """When converting a tag in this group directly to a string, include units
        or not.
        """
        self._str_units = value

    @property
    def set_in_display_units(self):
        """When this is True, and set() or fix() are called on a tag in the group,
        with a value that doesn't include units, assume the display units.
        """
        return self._set_in_display_units

    @set_in_display_units.setter
    def set_in_display_units(self, value):
        """When this is True, and set() or fix() are called on a tag in the group,
        with a value that doesn't include units, assume the display units.
        """
        self._set_in_display_units = value


def svg_tag(
    svg=None,
    tag_group=None,
    outfile=None,
    idx=None,
    tag_map=None,
    show_tags=False,
    byte_encoding="utf-8",
):
    """
    Replace text in a SVG with tag values for the model. This works by looking
    for text elements in the SVG with IDs that match the tags or are in tag_map.

    Args:
        svg: a file pointer or a string containing svg contents
        tag_group: a ModelTagGroup with tags to display in the SVG
        outfile: a file name to save the results, if None don't save
        idx: if None not indexed, otherwise an index in the indexing set of the
            reference
        tag_map: dictionary with svg id keys and tag values, to map svg ids to
            tags, used in cases where tags contain characters that cannot be used
            in the svg's xml
        show_tags: Put tag labels of the diagram instead of numbers
        byte_encoding: If svg is given as a byte-array, use this encoding to
            convert it to a string.

    Returns:
        SVG String
    """
    if svg is None:
        raise RuntimeError("svg string or file-like object required.")

    # get SVG content string
    if isinstance(svg, str):  # already a string
        pass
    elif isinstance(svg, bytes):  # bytes to string
        svg = svg.decode(byte_encoding)  # file-like to string
    elif hasattr(svg, "read"):
        svg = svg.read()
    else:  # Can't handle whatever this is.
        raise TypeError("SVG must either be a string or a file-like object")

    # Make tag map here because the tags may not make valid XML IDs if no
    # tag_map provided we'll go ahead and handle XML @ (maybe more in future)
    if tag_map is None:
        tag_map = dict()
        for tag in tag_group:
            new_tag = tag.replace("@", "_")
            new_tag = new_tag.replace(" ", "_")
            tag_map[new_tag] = tag

    # Turn SVG string into XML document
    doc = xml.dom.minidom.parseString(svg)
    # Get the text elements of the SVG
    texts = doc.getElementsByTagName("text")

    # Add some text
    for t in texts:
        _id = t.attributes["id"].value
        if _id in tag_map:
            # if it's multiline change last line
            try:
                tspan = t.getElementsByTagName("tspan")[-1]
            except IndexError:
                _log.warning(f"Text object but no tspan for tag {tag_map[_id]}.")
                _log.warning(f"Skipping output for {tag_map[_id]}.")
                continue
            try:
                tspan = tspan.childNodes[0]
            except IndexError:
                # No child node means there is a line with no text, so add some.
                tspan.appendChild(doc.createTextNode(""))
                tspan = tspan.childNodes[0]

            if show_tags:
                val = tag_map[_id]
            else:
                if tag_group[tag_map[_id]].is_indexed:
                    val = tag_group[tag_map[_id]][idx]
                else:
                    val = tag_group[tag_map[_id]]

            tspan.nodeValue = str(val)

    new_svg = doc.toxml()
    # If outfile is provided save to a file
    if outfile is not None:
        with open(outfile, "w") as f:
            f.write(new_svg)
    # Return the SVG as a string.  This lets you take several passes at adding
    # output without saving and loading files.
    return new_svg



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import os

import pytest
from pyomo.environ import SolverFactory

from idaes.core.util.testing import _enable_scip_solver_for_testing


ampl_module_scip = pytest.importorskip(
    "ampl_module_scip", reason="'ampl_module_scip' not available"
)


@pytest.mark.unit
def test_path_manipulation():
    path_before_enabling = str(os.environ["PATH"])

    func_to_revert_changes = _enable_scip_solver_for_testing()
    path_after_enabling = str(os.environ["PATH"])
    sf = SolverFactory("scip")
    assert len(path_after_enabling) > len(path_before_enabling)
    assert str(ampl_module_scip.bin_dir) in str(sf.executable())
    assert sf.available()

    func_to_revert_changes()
    path_after_reverting = str(os.environ["PATH"])
    assert path_after_reverting == path_before_enabling



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
# TODO: Missing doc strings
# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring

import pandas as pd


class OffsetScaler(object):
    @staticmethod
    def create_normalizing_scaler(dataframe):
        """
        Creates a scaling object that normalizes the data between 0 and 1

        Args:
           dataframe: pandas DataFrame
              The dataframe containing the data (usually the training data)
              that will be used to compute the scaling factor and offset
        """
        expected_columns = list(dataframe.columns)
        offset = dataframe.min()
        factor = dataframe.max() - dataframe.min()
        return OffsetScaler(expected_columns, offset, factor)

    @staticmethod
    def create_from_mean_std(dataframe):
        """
        Creates a scaling object using the mean as the offset and the
        standard deviation as the as the factor

        Args:
           dataframe: pandas DataFrame
              The dataframe containing the data (usually the training data) that will
              be used to compute the mean and standard deviation for the scaler
        """
        expected_columns = list(dataframe.columns)
        offset = dataframe.mean()
        factor = dataframe.std()
        return OffsetScaler(expected_columns, offset, factor)

    def __init__(self, expected_columns, offset_series, factor_series):
        """
        This scaling object shifts by the offset and then scales the result
        by the factor. Typically, one would create this with the static method
        create_from_dataframe.

        scaled_data = (data-offset)/factor

        Args:
           expected_columns: list of str
              list of strings indicating the names of the columns in offset_series,
              factor_series, and the dataframe passed to scale and unscale.
           offset_series: pandas Series
              Series with columns (or labels) the same as expected_columns and
              values that represent the offset to be used when shifting the data
           factor_series: pandas Series
              Series with columns (or labels) the same as expected_columns and
              values that represent the factor to be used to scale the shifted data
        """
        self._expected_columns = expected_columns
        self._offset = offset_series
        self._factor = factor_series
        if list(offset_series.index) != expected_columns:
            raise ValueError(
                "OffsetScaler was passed an offset series with an index that"
                " does not match expected_columns. Please make sure these labels match."
            )
        if list(factor_series.index) != expected_columns:
            raise ValueError(
                "OffsetScaler was passed a factor series with an index that"
                " does not match expected_columns. Please make sure these labels match."
            )

    def _verify_columns_match(self, dataframe):
        if self._expected_columns != list(dataframe.columns):
            raise ValueError(
                "OffsetScaler was passed a dataframe that did not contain"
                " the same column labels as those used to create the scaler."
                " Please make sure the column labels match."
            )

    def scale(self, dataframe):
        """
        Return a new dataframe where the values are scaled according to the
        offset and factor

        Args:
           dataframe: pandas Dataframe
              The dataframe to be scaled

        Returns: pandas DataFrame
        """
        self._verify_columns_match(dataframe)
        df = dataframe - self._offset
        df = df.divide(self._factor)
        return df

    def unscale(self, dataframe):
        """
        Return a new dataframe where the values are unscaled according to the
        offset and factor

        Args:
           dataframe: pandas Dataframe
              The dataframe to be unscaled

        Returns: pandas DataFrame
        """
        self._verify_columns_match(dataframe)
        df = dataframe.multiply(self._factor)
        df = df + self._offset
        return df

    def expected_columns(self):
        """
        Return the expected column names for the scaler series objects
        """
        return self._expected_columns

    def offset_series(self):
        """
        Return the offset for the scaler as a pandas Series object
        """
        return self._offset

    def factor_series(self):
        """
        Return the factors for the scaler as a pandas Series object
        """
        return self._factor

    def to_dict(self):
        """
        Returns a dictionary representation of this scaler
        """
        d = dict()
        d["expected_columns"] = list(self._expected_columns)
        d["offset"] = self._offset.to_dict()
        d["factor"] = self._factor.to_dict()
        return d

    @staticmethod
    def from_dict(d):
        """
        Create an instance of this scaler from a dictionary
        (that was created with to_dict)

        Args:
           d : dict
              The dict created with to_dict

        Returns: new OffsetScaler
        """
        expected_columns = d["expected_columns"]
        offset = pd.Series(d["offset"])
        factor = pd.Series(d["factor"])
        return OffsetScaler(expected_columns, offset, factor)
